#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test `corrct.regularizers` package.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np

import pytest

from corrct import regularizers
from corrct import testing


eps = np.finfo(np.float32).eps


@pytest.fixture(scope="class")
def bootstrap_base(request):
    test_vols_shape = (100, 100, 10)
    cls = request.cls
    cls.vol_rand_2d = np.fmin(np.random.rand(*test_vols_shape[:2]) + eps, 1)
    cls.vol_rand_3d = np.fmin(np.random.rand(*test_vols_shape) + eps, 1)


@pytest.mark.usefixtures("bootstrap_base")
class TestRegularizers:
    """Tests for the regularizers in `corrct.regularizers` package."""

    def _test_Regularizer_l1(self, vol):
        weight = 0.5
        reg = regularizers.Regularizer_l1(weight)

        tau = reg.initialize_sigma_tau(vol)
        assert tau == 1

        dual = reg.initialize_dual()
        assert np.all(dual.shape == vol.shape)
        assert np.all(dual == 0)

        copy_dual = dual.copy()
        reg.update_dual(dual, vol)
        assert not np.all(dual == copy_dual)

        assert np.all(dual == vol)

        dual += weight / 2
        copy_dual = dual.copy()
        reg.apply_proximal(dual)
        assert np.all(np.isclose(dual, np.fmin(weight, copy_dual)))

        dual = copy_dual - weight * 2
        copy_dual = dual.copy()
        reg.apply_proximal(dual)
        assert np.all(np.isclose(dual, np.fmax(-weight, copy_dual)))

        upd = reg.compute_update_primal(dual)
        assert np.all(np.isclose(dual, upd))

    def _test_Regularizer_l1swl(self, vol):
        weight = 0.5
        ndims = len(vol.shape)
        level = 2
        reg = regularizers.Regularizer_l1swl(weight, "db1", level, ndims=ndims, min_approx=True)

        reg.initialize_sigma_tau(vol)

        dual = reg.initialize_dual()
        assert np.all(dual.shape[1:] == testing.roundup_to_pow2(vol.shape, level))
        assert dual.shape[0] == ((2**ndims - 1) * level + 1)
        assert np.all(dual == 0)

        copy_dual = dual.copy()
        reg.update_dual(dual, vol)
        sigma_shape = [-1] + [1] * (len(reg.op.adj_shape) - 1)
        assert np.all(np.isclose(dual, copy_dual + reg.op(vol) * np.reshape(reg.sigma, sigma_shape)))

        upd = reg.compute_update_primal(dual)
        assert np.all(np.isclose(upd, reg.op.T(dual)))

        copy_dual = dual.copy()
        reg.apply_proximal(dual)
        assert np.all(np.isclose(dual, np.fmax(np.fmin(weight, copy_dual), -weight)))

    def _test_Regularizer_TV(self, vol):
        weight = 0.5
        ndims = len(vol.shape)
        reg = regularizers.Regularizer_Grad(weight, ndims=ndims)

        tau = reg.initialize_sigma_tau(vol)
        assert tau == (2 * ndims)

        dual = reg.initialize_dual()
        assert np.all(dual.shape[1:] == vol.shape)
        assert dual.shape[0] == ndims
        assert np.all(dual == 0)

        copy_dual = dual.copy()
        reg.update_dual(dual, vol)
        assert np.all(np.isclose(dual, copy_dual + reg.op(vol) / 2))

        dual = np.random.random(dual.shape)
        upd = reg.compute_update_primal(dual)
        assert np.all(np.isclose(reg.op.T(dual), upd))

    def _test_Regularizer_lap(self, vol):
        weight = 0.25
        ndims = len(vol.shape)
        reg = regularizers.Regularizer_lap(weight, ndims=ndims)

        tau = reg.initialize_sigma_tau(vol)
        assert tau == (4 * ndims)

        dual = reg.initialize_dual()
        assert np.all(dual.shape == vol.shape)
        assert np.all(dual == 0)

        copy_dual = dual.copy()
        reg.update_dual(dual, vol)
        assert np.all(np.isclose(dual, copy_dual + reg.op(vol) / 4))

        dual = np.random.random(dual.shape)
        upd = reg.compute_update_primal(dual)
        assert np.all(np.isclose(reg.op.T(dual), upd))

    def test_000_Regularizer_l1_2D(self):
        """Test l1-min regularizer in 2D."""
        self._test_Regularizer_l1(self.vol_rand_2d)

    def test_001_Regularizer_l1_3D(self):
        """Test l1-min regularizer in 3D."""
        self._test_Regularizer_l1(self.vol_rand_3d)

    def test_002_Regularizer_l1swl_2D(self):
        """Test l1-min wavelet regularizer in 2D."""
        self._test_Regularizer_l1swl(self.vol_rand_2d)

    def test_003_Regularizer_l1swl_3D(self):
        """Test l1-min wavelet regularizer in 3D."""
        self._test_Regularizer_l1swl(self.vol_rand_3d)

    def test_004_Regularizer_TV_2D(self):
        """Test TV regularizer in 2D."""
        self._test_Regularizer_TV(self.vol_rand_2d)

    def test_005_Regularizer_TV_3D(self):
        """Test TV regularizer in 3D."""
        self._test_Regularizer_TV(self.vol_rand_3d)

    def test_004_Regularizer_lap_2D(self):
        """Test l1-min laplacian regularizer in 2D."""
        self._test_Regularizer_lap(self.vol_rand_2d)

    def test_005_Regularizer_lap_3D(self):
        """Test l1-min laplacian regularizer in 3D."""
        self._test_Regularizer_lap(self.vol_rand_3d)



#!/usr/bin/env python3
"""
xraylib handling functions.

@author: Nicola VIGANÒ, CEA-IRIG, Grenoble, France
"""

from typing import Union

import numpy as np

import xraylib

xraylib.XRayInit()


def get_element_number(element: Union[str, int]) -> int:
    """Return the element number from the symbol.

    Parameters
    ----------
    element : str | int
        The element symbol (or number, which won't be converted).

    Returns
    -------
    int
        The corresponding element number.
    """
    if isinstance(element, int):
        return element
    else:
        return xraylib.SymbolToAtomicNumber(element)


def get_element_number_and_symbol(element: Union[str, int]) -> tuple[str, int]:
    """Return both the element symbol and number from either the symbol or the number.

    Parameters
    ----------
    element : str | int
        The element symbol (or number, which won't be converted).

    Returns
    -------
    tuple[str, int]
        The element symbol and number.
    """
    if isinstance(element, str):
        el_sym = element
        el_num = xraylib.SymbolToAtomicNumber(element)
    else:
        el_sym = xraylib.AtomicNumberToSymbol(element)
        el_num = element

    return el_sym, el_num


def get_compound(cmp_name: str, density: Union[float, None] = None) -> dict:
    """
    Build a compound from the compound composition string.

    Parameters
    ----------
    cmp_name : str
        Compund name / composition.
    density : float, optional
        The density of the compound. If not provided it will be approximated from the composing elements.
        The default is None.

    Returns
    -------
    cmp : dict
        The compound structure.
    """
    try:
        cmp = xraylib.GetCompoundDataNISTByName(cmp_name)
    except ValueError:
        cmp = xraylib.CompoundParser(cmp_name)

    cmp["name"] = cmp_name
    if density is None:
        density = 0
        for ii, el in enumerate(cmp["Elements"]):
            density += xraylib.ElementDensity(el) * cmp["massFractions"][ii]
    cmp["density"] = density
    return cmp


def get_compound_cross_section(compound: dict, mean_energy_keV: float) -> float:
    """Compute a compound cross section for the given incoming photon energy.

    Parameters
    ----------
    compound : dict
        The compound structure (as returned by `get_compound`)
    mean_energy_keV : float
        The average photon energy

    Returns
    -------
    float
        The computed cross-section
    """
    try:
        return xraylib.CS_Total_CP(compound["name"], mean_energy_keV)
    except ValueError:
        elemets_cs = [
            xraylib.CS_Total(el, mean_energy_keV) * compound["massFractions"][ii] for ii, el in enumerate(compound["Elements"])
        ]
        return np.sum(elemets_cs, axis=0)



"""
Miscellaneous processing routines.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np
import scipy as sp

from typing import Optional, Union, Callable
from collections.abc import Sequence
from numpy.typing import NDArray, ArrayLike, DTypeLike

import matplotlib.pyplot as plt


eps = np.finfo(np.float32).eps

NDArrayInt = NDArray[np.signedinteger]


def circular_mask(
    vol_shape_zxy: Union[Sequence[int], NDArrayInt],
    radius_offset: float = 0,
    coords_ball: Union[Sequence[int], NDArrayInt, None] = None,
    ball_norm: float = 2,
    vol_origin_zxy: Union[Sequence[float], NDArray, None] = None,
    taper_func: Optional[str] = None,
    taper_target: Union[str, float] = "edge",
    super_sampling: int = 1,
    squeeze: bool = True,
    dtype: DTypeLike = np.float32,
) -> NDArray:
    """
    Compute a circular mask for the reconstruction volume.

    Parameters
    ----------
    vol_shape_zxy : Sequence[int] | NDArrayInt
        The size of the volume.
    radius_offset : float, optional
        The offset with respect to the volume edge. The default is 0.
    coords_ball : Sequence[int] | NDArrayInt | None, optional
        The coordinates to consider for the non-masked region. The default is None.
    ball_norm : float, optional
        The norm of the ball. The default is 2.
    vol_origin_zxy : Optional[Sequence[float]], optional
        The origin of the coordinates in voxels. The default is None.
    taper_func : str, optional
        The mask data type. Allowed types: "const" | "cos". The default is "const".
    taper_target : str | float, optional
        The size target. Allowed values: "edge" | "diagonal". The default is "edge".
    super_sampling : int, optional
        The pixel super sampling to be used for the mask. The default is 1.
    squeeze : bool, optional
        Whether to squeeze the mask. The default is True.
    dtype : DTypeLike, optional
        The type of mask. The default is np.float32.

    Raises
    ------
    ValueError
        In case of unknown taper_func value, or mismatching volume origin and shape.

    Returns
    -------
    NDArray
        The circular mask.
    """
    vol_shape_zxy_s = np.array(vol_shape_zxy, dtype=int) * super_sampling

    coords = [
        np.linspace(-(s - 1) / (2 * super_sampling), (s - 1) / (2 * super_sampling), s, dtype=dtype) for s in vol_shape_zxy_s
    ]
    if vol_origin_zxy is not None:
        if len(coords) != len(vol_origin_zxy):
            raise ValueError(f"The volume shape ({len(coords)}), and the origin shape ({len(vol_origin_zxy)}) should match")
        coords = [c + vol_origin_zxy[ii] for ii, c in enumerate(coords)]
    coords = np.meshgrid(*coords, indexing="ij")

    if coords_ball is None:
        coords_ball = np.arange(-np.fmin(2, len(vol_shape_zxy_s)), 0, dtype=int)
    else:
        coords_ball = np.array(coords_ball, dtype=int)

    max_radius = np.min(vol_shape_zxy_s[coords_ball]) / (2 * super_sampling) + radius_offset

    coords = np.stack(coords, axis=0)
    if coords_ball.size == 1:
        dists = np.abs(coords[coords_ball, ...])
    else:
        dists = np.linalg.norm(coords[coords_ball, ...], axis=0, ord=ball_norm)

    if taper_func is None:
        mask = (dists <= max_radius).astype(dtype)
    elif isinstance(taper_func, str):
        if isinstance(taper_target, str):
            if taper_target.lower() == "edge":
                cut_off_denom = 2
                cut_off_offset = 0
            elif taper_target.lower() == "diagonal":
                cut_off_denom = np.sqrt(2)
                cut_off_offset = 0
            else:
                raise ValueError(
                    f"When `taper_target` is str, it should be one of: 'edge' | 'diagonal', but {taper_target} passed instead."
                )
        else:
            cut_off_denom = 2
            if taper_target < radius_offset:
                print(f"WARNING: parameter `taper_target`={taper_target} is smaller than `radius_offset`={radius_offset}.")
            cut_off_offset = np.fmax(taper_target, radius_offset)

        if taper_func.lower() == "cos":
            cut_off_radius = np.min(vol_shape_zxy_s[coords_ball]) / (cut_off_denom * super_sampling) + cut_off_offset
            cut_off_size = cut_off_radius - max_radius
            outter_vals = np.cos(np.fmax(dists - max_radius, 0) / cut_off_size * np.pi) / 2 + 0.5
            mask = (outter_vals * (dists < cut_off_radius)).astype(dtype)
        else:
            raise ValueError(f"Unknown taper function: {taper_func}")
    else:
        raise ValueError(f"Parameter `taper_func` should either be a string or None.")

    if super_sampling > 1:
        new_shape = np.stack([np.array(vol_shape_zxy), np.ones_like(vol_shape_zxy) * super_sampling], axis=1).flatten()
        mask = mask.reshape(new_shape)
        mask = np.mean(mask, axis=tuple(np.arange(1, len(vol_shape_zxy) * 2, 2, dtype=int)))

    if squeeze:
        mask = np.squeeze(mask)

    return mask


def ball(
    data_shape_vu: ArrayLike,
    radius: Union[int, float],
    super_sampling: int = 5,
    dtype: DTypeLike = np.float32,
    func: Optional[Callable] = None,
) -> ArrayLike:
    """
    Compute a ball with specified radius.

    Parameters
    ----------
    data_shape_vu : ArrayLike
        Shape of the output array.
    radius : int | float
        Radius of the ball.
    super_sampling : int, optional
        Super-sampling for having smoother ball edges. The default is 5.
    dtype : DTypeLike, optional
        Type of the output. The default is np.float32.
    func : Optional[Callable], optional
        Point-wise function for the local values. The default is None.

    Returns
    -------
    ArrayLike
        The ball.
    """
    data_shape_vu = np.array(data_shape_vu, dtype=int) * super_sampling

    # coords = [np.linspace(-(s - 1) / 2, (s - 1) / 2, s, dtype=np.float32) for s in data_shape_vu]
    coords = [np.fft.fftfreq(d, 1 / d) for d in data_shape_vu]
    coords = np.stack(np.meshgrid(*coords, indexing="ij"), axis=0)

    r = np.sqrt(np.sum(coords**2, axis=0)) / super_sampling

    probe = (r < radius).astype(dtype)
    if func is not None:
        probe *= func(r)

    probe = np.roll(probe, super_sampling // 2, axis=tuple(np.arange(len(data_shape_vu))))
    new_shape = np.stack([data_shape_vu // super_sampling, np.ones_like(data_shape_vu) * super_sampling], axis=1).flatten()
    probe = probe.reshape(new_shape)
    probe = np.mean(probe, axis=tuple(np.arange(1, len(data_shape_vu) * 2, 2, dtype=int)))

    return np.fft.fftshift(probe)


def azimuthal_integration(img: NDArray, axes: Sequence[int] = (-2, -1), domain: str = "direct") -> NDArray:
    """
    Compute the azimuthal integration of a n-dimensional image or a stack of them.

    Parameters
    ----------
    img : NDArray
        The image or stack of images.
    axes : tuple(int, int), optional
        Axes of that need to be azimuthally integrated. The default is (-2, -1).
    domain : string, optional
        Domain of the integration. Options are: "direct" | "fourier". Default is "direct".

    Raises
    ------
    ValueError
        Error returned when not passing images or wrong axes.

    Returns
    -------
    NDArray
        The azimuthally integrated profile.
    """
    num_dims_int = len(axes)
    num_dims_img = len(img.shape)

    if num_dims_img < num_dims_int:
        raise ValueError(
            "Input image ({num_dims_img}D) should be at least the same dimensionality"
            " of the axes for the integration (#{num_dims_int})."
        )
    if len(axes) == 0:
        raise ValueError("Input axes should be at least 1.")

    # Compute the coordinates of the pixels along the chosen axes
    img_axes_dims = np.array(np.array(img.shape)[list(axes)], ndmin=1)
    if domain.lower() == "direct":
        half_dims = (img_axes_dims - 1) / 2
        coords = [np.linspace(-h, h, d) for h, d in zip(half_dims, img_axes_dims)]
    else:
        coords = [np.fft.fftfreq(d, 1 / d) for d in img_axes_dims]
    coords = np.stack(np.meshgrid(*coords, indexing="ij"))
    r = np.sqrt(np.sum(coords**2, axis=0))

    # Reshape the volume to have the axes to be integrates as right-most axes
    img_tr_op = np.array([*range(len(img.shape))])
    img_tr_op = np.concatenate((np.delete(img_tr_op, obj=axes), img_tr_op[list(axes)]))
    img = np.transpose(img, img_tr_op)

    if num_dims_img > num_dims_int:
        img_old_shape = img.shape[:-num_dims_int]
        img = np.reshape(img, [-1, *img_axes_dims])

    # Compute the linear interpolation coefficients
    r_l = np.floor(r)
    r_u = r_l + 1
    w_l = (r_u - r) * img
    w_u = (r - r_l) * img

    # Do the azimuthal integration as a histogram operation
    r_all = np.concatenate((r_l.flatten(), r_u.flatten())).astype(int)
    if num_dims_img > num_dims_int:
        num_imgs = img.shape[0]
        az_img = []
        for ii in range(num_imgs):
            w_all = np.concatenate((w_l[ii, ...].flatten(), w_u[ii, ...].flatten()))
            az_img.append(np.bincount(r_all, weights=w_all))
        az_img = np.array(az_img)
        return np.reshape(az_img, (*img_old_shape, az_img.shape[-1]))  # type: ignore
    else:
        w_all = np.concatenate((w_l.flatten(), w_u.flatten()))
        return np.bincount(r_all, weights=w_all)


def lines_intersection(
    line_1: NDArray,
    line_2: Union[float, NDArray],
    position: str = "first",
    x_lims: Optional[tuple[Optional[float], Optional[float]]] = None,
) -> Optional[tuple[float, float]]:
    """
    Compute the intersection point between two lines.

    Parameters
    ----------
    line_1 : NDArray
        The first line.
    line_2 : float | NDArray
        The second line. It can be a scalar representing a horizontal line.
    position : str, optional
        The position of the point to select. Either "first" or "last".
        The default is "first".

    Raises
    ------
    ValueError
        If position is neither "first" nor "last".

    Returns
    -------
    Tuple[float, float] | None
        It returns either the requested crossing point, or None in case the
        point was not found.
    """
    line_1 = np.array(np.squeeze(line_1), ndmin=1)
    line_2 = np.array(np.squeeze(line_2), ndmin=1)
    # Find the transition points, by first finding where line_2 is above line_1
    crossing_points = np.where(line_2 > line_1, 0, 1)
    crossing_points = np.abs(np.diff(crossing_points))

    if x_lims is not None:
        if x_lims[0] is None:
            if x_lims[1] is None:
                raise ValueError("When passing `x_lims`, at least one of the values should not be None.")
            else:
                bias = 0
                crossing_points = crossing_points[: x_lims[1]]
        else:
            bias = x_lims[0]
            if x_lims[1] is None:
                crossing_points = crossing_points[x_lims[0] :]
            else:
                crossing_points = crossing_points[x_lims[0] : x_lims[1]]
    else:
        bias = 0

    crossing_points = np.where(crossing_points)[0]

    if crossing_points.size == 0:
        print("No crossing found!")
        return None

    if position.lower() == "first":
        point_l = crossing_points[0] + bias
    elif position.lower() == "last":
        point_l = crossing_points[-1] + bias
    else:
        raise ValueError(f"Crossing position: {position} unknown. Please choose either 'first' or 'last'.")

    x1 = 0.0
    x2 = 1.0
    y1 = line_1[point_l]
    y2 = line_1[point_l + 1]
    x3 = 0.0
    x4 = 1.0
    if line_2.size == 1:
        y3 = line_2
        y4 = line_2
    else:
        y3 = line_2[point_l]
        y4 = line_2[point_l + 1]

    # From wikipedia: https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_two_points_on_each_line
    p_den = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    p_x_num = (x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)
    p_y_num = (x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)

    p_x = p_x_num / p_den + point_l
    p_y = p_y_num / p_den

    return float(p_x), float(p_y)


def norm_cross_corr(
    img1: NDArray,
    img2: Optional[NDArray] = None,
    axes: Sequence[int] = (-2, -1),
    t_match: bool = False,
    mode_full: bool = True,
    compute_profile: bool = True,
    plot: bool = True,
) -> Union[NDArray, tuple[NDArray, NDArray]]:
    """
    Compute the normalized cross-correlation between two images.

    Parameters
    ----------
    img1 : NDArray
        The first image.
    img2 : NDArray, optional
        The second images. If None, it computes the auto-correlation. The default is None.
    axes : Sequence[int], optional
        Axes along which to compute the cross-correlation. The default is (-2, -1).
    t_match : bool, optional
        Whether to perform the cross-correlation for template matching. The default is False.
    mode_full : bool, optional
        Whether to return the "full" or "same" convolution size. The default is True.
    compute_profile : bool, optional
        Whether to compute the azimuthal integration of the cross-correlation or not. The default is True.
    plot : bool, optional
        Whether to plot the profile of the cross-correlation curve. The default is True.

    Returns
    -------
    NDArray
        The one-dimensional cross-correlation curve.
    """

    def local_sum(x: NDArray, axes: Sequence[int]) -> NDArray:
        padding = np.zeros(len(x.shape), dtype=int)
        for a in axes:
            padding[a] = x.shape[a]
        y: NDArray[np.floating] = np.pad(x, padding)
        for a in axes:
            y = np.cumsum(y, axis=a)
            slicing1 = [slice(None)] * len(x.shape)
            slicing1[a] = slice(x.shape[a], -1)
            slicing2 = [slice(None)] * len(x.shape)
            slicing2[a] = slice(0, -x.shape[a] - 1)
            y = y[tuple(slicing1)] - y[tuple(slicing2)]
        return y

    if img2 is None:
        is_autocorrelation = True
        img2 = img1
    else:
        is_autocorrelation = False

    img1 = img1.astype(np.float32)
    img2 = img2.astype(np.float32)

    for a in axes:
        img2 = np.flip(img2, axis=a)
    cc = sp.signal.fftconvolve(img1, img2, mode="full", axes=axes)

    if not mode_full:
        slices = [slice(None)] * len(cc.shape)
        for a in axes:
            start_ind = (cc.shape[a] - img1.shape[a]) // 2
            end_ind = start_ind + img1.shape[a]
            slices[a] = slice(start_ind, end_ind)
        slices = tuple(slices)
        cc = cc[slices]

    if t_match:
        local_sums_img2 = local_sum(img2, axes=axes)
        local_sums_img2_2 = local_sum(img2**2, axes=axes)
        if not mode_full:
            local_sums_img2 = local_sums_img2[slices]
            local_sums_img2_2 = local_sums_img2_2[slices]

        cc_n = cc - local_sums_img2 * np.mean(img1)

        cc_n /= np.std(img1) * np.sqrt(np.prod(np.array(img1.shape)[list(axes)]))

        diff_local_sums = local_sums_img2_2 - (local_sums_img2**2) / np.prod(np.array(img2.shape)[list(axes)])
        cc_n /= np.sqrt(diff_local_sums.clip(0, None))
    else:
        cc_n = cc / (np.linalg.norm(img1) * np.linalg.norm(img2))

    cc_n = np.fft.ifftshift(cc_n)

    if compute_profile:
        cc_l = azimuthal_integration(cc_n, axes=axes, domain="fourier")
        cc_o = azimuthal_integration(np.ones_like(cc_n), axes=axes, domain="fourier")

        cc_l /= cc_o
        cc_l = cc_l[: np.min(np.array(img1.shape)[list(axes)])]

        if plot:
            p_xy = lines_intersection(cc_l, 0.5, position="first")

            fig, axs = plt.subplots(1, 1, figsize=(7, 3.5))
            axs.plot(cc_l, label="Auto-correlation" if is_autocorrelation else "Cross-correlation")
            axs.plot(np.ones_like(cc_l) * 0.5, label="Half-maximum")
            if p_xy is not None:
                axs.scatter(p_xy[0], p_xy[1])
                axs.plot([p_xy[0], p_xy[0]], [0, 1], label=f"Resolution: {p_xy[0]:.3} pix")
            axs.grid()
            axs.legend(fontsize=13)
            axs.tick_params(labelsize=16)
            axs.set_title("Cross-correlation")
            fig.tight_layout()
            plt.show(block=False)

        return cc_n, cc_l
    else:
        return cc_n


def inspect_fourier_img(img: NDArray, remove_zero: bool = False) -> None:
    """Display Fourier representation of the input image.

    Parameters
    ----------
    img : NDArray
        Input image.
    remove_zero : bool, optional
        Remove the zero frequency value. The default is False.
    """
    img_f = np.squeeze(np.fft.fft2(img))
    if remove_zero is True:
        img_f[0, 0] = 0
    img_f_sh = np.fft.fftshift(img_f)

    f, axs = plt.subplots(2, 3)
    f.suptitle("Fourier representation")

    axs[0, 0].imshow(np.real(img_f_sh))
    axs[0, 0].set_title("Real")
    axs[0, 1].imshow(np.imag(img_f_sh))
    axs[0, 1].set_title("Imag")
    axs[0, 2].imshow(np.abs(img_f_sh))
    axs[0, 2].set_title("Abs")

    vert_img_f = np.fft.fftshift(img_f[:, 0])
    axs[1, 0].plot(np.stack((np.real(vert_img_f), np.imag(vert_img_f), np.abs(vert_img_f)), axis=1))
    axs[1, 0].set_title("Vertical profiles")

    horz_img_f = np.fft.fftshift(img_f[0, :])
    axs[1, 1].plot(np.stack((np.real(horz_img_f), np.imag(horz_img_f), np.abs(horz_img_f)), axis=1))
    axs[1, 1].set_title("Horizontal profiles")

    diag_img_f = np.fft.fftshift(np.diag(img_f))
    axs[1, 2].plot(np.stack((np.real(diag_img_f), np.imag(diag_img_f), np.abs(diag_img_f)), axis=1))
    axs[1, 2].set_title("Diagonal profiles")

    plt.show(block=False)



"""
Incident beam and emidded radiation attenuation support.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import concurrent.futures as cf
import multiprocessing as mp
from collections.abc import Sequence
from typing import Callable, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes._axes import Axes
from numpy.typing import ArrayLike, DTypeLike, NDArray
from tqdm.auto import tqdm

from corrct import _projector_backends as prj_backends
from corrct import models
from corrct.physics.xraylib_helper import get_compound, get_compound_cross_section

num_threads = round(np.log2(mp.cpu_count() + 1))

NDArrayFloat = NDArray[np.floating]
NDArrayInt = NDArray[np.integer]

CONVERT_UM_TO_CM = 1e-4


class AttenuationVolume:
    """Attenuation volume computation class."""

    incident_local: Union[NDArrayFloat, None]
    emitted_local: Union[NDArrayFloat, None]
    angles_rot_rad: NDArrayFloat
    angles_det_rad: NDArrayFloat

    dtype: DTypeLike

    vol_shape_zyx: NDArray
    maps: NDArray

    def __init__(
        self,
        incident_local: Union[NDArrayFloat, None],
        emitted_local: Union[NDArrayFloat, None],
        angles_rot_rad: ArrayLike,
        angles_det_rad: Union[NDArrayFloat, ArrayLike, float] = np.pi / 2,
        dtype: DTypeLike = np.float32,
    ):
        """
        Initialize the AttenuationVolume class.

        Raises
        ------
        ValueError
            In case no volumes were passed, or if they differed in shape.
        """
        self.incident_local = incident_local
        self.emitted_local = emitted_local
        self.angles_rot_rad = np.array(angles_rot_rad, ndmin=1)
        self.angles_det_rad = np.array(angles_det_rad, ndmin=1)

        self.dtype = dtype

        if self.incident_local is not None:
            self.vol_shape_zyx = np.array(self.incident_local.shape)

            if self.emitted_local is not None and np.any(self.vol_shape_zyx != self.emitted_local.shape):
                raise ValueError(
                    f"Incident volume shape ({self.incident_local.shape}) does not"
                    + f" match the emitted volume shape ({self.emitted_local.shape})"
                )
        elif self.emitted_local is not None:
            self.vol_shape_zyx = np.array(self.emitted_local.shape)
        else:
            raise ValueError("No attenuation volumes were given.")

        self.vol_shape_zyx = np.array(self.vol_shape_zyx, ndmin=1)

        num_dims = len(self.vol_shape_zyx)
        if num_dims not in [2, 3]:
            raise ValueError(f"Maps can only be 2D or 3D Arrays. A {num_dims}-dimensional was passed ({self.vol_shape_zyx}).")

    def _compute_attenuation_angle_in(self, local_att: NDArrayFloat, angle_rad: float) -> NDArray:
        return prj_backends.compute_attenuation(local_att, angle_rad, invert=False)[None, ...]

    def _compute_attenuation_angle_out(self, local_att: NDArrayFloat, angle_rad: float) -> NDArray:
        angle_det = angle_rad + self.angles_det_rad
        atts = np.empty(self.maps.shape[1:], dtype=self.dtype)
        for ii, a in enumerate(angle_det):
            atts[ii, ...] = prj_backends.compute_attenuation(local_att, a, invert=True)
        return atts

    def compute_maps(self, use_multithreading: bool = True, verbose: bool = True) -> None:
        """
        Compute the correction maps for each angle.

        Parameters
        ----------
        use_multithreading : bool, optional
            Use multi-threading for computing the attenuation maps. The default is True.
        verbose : bool, optional
            Show verbose output. The default is True.
        """
        num_rot_angles = len(self.angles_rot_rad)
        self.maps = np.ones([num_rot_angles, len(self.angles_det_rad), *self.vol_shape_zyx], dtype=self.dtype)

        def process_angles(
            func: Callable[[NDArray, float], NDArray], att_vol: NDArrayFloat, angles: NDArrayFloat, description: str
        ) -> None:
            if use_multithreading:
                with cf.ThreadPoolExecutor(max_workers=num_threads) as executor:
                    futures_to_angle = {executor.submit(func, att_vol, a): (ii, a) for ii, a in enumerate(angles)}
                    try:
                        for f in tqdm(
                            cf.as_completed(futures_to_angle),
                            desc=description,
                            disable=(not verbose),
                            total=num_rot_angles,
                        ):
                            ii, a = futures_to_angle[f]
                            try:
                                self.maps[ii, ...] *= f.result()
                            except ValueError as exc:
                                raise RuntimeError(f"Angle {a} (#{ii}) generated an exception") from exc
                    except:
                        print("Shutting down..", end="", flush=True)
                        executor.shutdown(cancel_futures=True)
                        print("\b\b: Done.")
                        raise
            else:
                for ii, a in enumerate(tqdm(angles, desc=description, disable=(not verbose))):
                    self.maps[ii, ...] *= func(att_vol, a)

        if self.incident_local is not None:
            description = "Computing attenuation maps for incident beam"
            process_angles(
                self._compute_attenuation_angle_in, self.incident_local, angles=self.angles_rot_rad, description=description
            )

        if self.emitted_local is not None:
            description = "Computing attenuation maps for emitted photons"
            process_angles(
                self._compute_attenuation_angle_out, self.emitted_local, angles=self.angles_rot_rad, description=description
            )

    def plot_map(
        self,
        ax: Axes,
        rot_ind: int,
        det_ind: int = 0,
        slice_ind: Optional[int] = None,
        axes: Union[Sequence[int], NDArrayInt] = (-2, -1),
    ) -> Sequence[float]:
        """
        Plot the requested attenuation map.

        Parameters
        ----------
        ax : matplotlib axes
            The axes where to plot.
        rot_ind : int
            Rotation angle index.
        det_ind : int, optional
            Detector angle index. The default is 0.
        slice_ind : Optional[int], optional
            Volume slice index (for 3D volumes). The default is None.
        axes : Sequence[int] | NDArray, optional
            Axes of the slice. The default is (-2, -1).

        Returns
        -------
        Sequence[float]
            The extent of the axes plot (min-max coords).

        Raises
        ------
        ValueError
            In case a slice index is not passed for a 3D volume.
        """
        att_map = np.squeeze(self.get_maps(rot_ind=rot_ind, det_ind=det_ind))
        other_dim = np.squeeze(np.delete(np.arange(-3, 0), axes))
        if len(att_map.shape) == 3:
            if slice_ind is None:
                raise ValueError("Slice index is needed for 3D volumes. None was passed.")

            att_map = np.take(att_map, slice_ind, axis=int(other_dim))

        slice_shape = self.vol_shape_zyx[list(axes)]
        coords = [(-(s - 1) / 2, (s - 1) / 2) for s in slice_shape]

        extent = tuple(np.concatenate(coords))
        ax.imshow(att_map, extent=extent)

        if other_dim == -3:
            arrow_length = np.linalg.norm(slice_shape) / np.pi
            arrow_args = dict(
                width=arrow_length / 25,
                head_width=arrow_length / 8,
                head_length=arrow_length / 6,
                length_includes_head=True,
            )

            prj_geom = models.ProjectionGeometry.get_default_parallel()
            beam_i_geom = prj_geom.rotate(-self.angles_rot_rad[rot_ind])
            beam_e_geom = prj_geom.rotate(-(self.angles_rot_rad[rot_ind] + self.angles_det_rad[det_ind]))

            beam_i_dir = beam_i_geom.src_pos_xyz[0, :2] * arrow_length
            beam_i_orig = -beam_i_dir
            beam_e_dir = beam_e_geom.src_pos_xyz[0, :2] * arrow_length
            beam_e_orig = np.array([0, 0])

            ax.arrow(*beam_i_orig, *beam_i_dir, **arrow_args, fc="r", ec="r")
            ax.arrow(*beam_e_orig, *beam_e_dir, **arrow_args, fc="k", ec="k")

        return extent

    def get_maps(
        self,
        roi: Optional[ArrayLike] = None,
        rot_ind: Union[int, slice, Sequence[int], NDArrayInt, None] = None,
        det_ind: Union[int, slice, Sequence[int], NDArrayInt, None] = None,
    ) -> NDArray:
        """
        Return the attenuation maps.

        Parameters
        ----------
        roi : ArrayLike, optional
            The region-of-interest to select. The default is None.
        rot_ind : int, optional
            A specific rotation index, if only one is to be desired. The default is None.
        det_ind : int, optional
            A specific detector index, if only one is to be desired. The default is None.

        Returns
        -------
        NDArray
            The attenuation maps.
        """
        maps = self.maps

        if rot_ind is not None:
            if isinstance(rot_ind, int):
                rot_ind = slice(rot_ind, rot_ind + 1, 1)
            maps = maps[rot_ind, ...]

        if det_ind is not None:
            if isinstance(det_ind, int):
                det_ind = slice(det_ind, det_ind + 1, 1)
            maps = maps[:, det_ind, ...]

        if roi is not None:
            raise NotImplementedError("Extracting a region of interest is not supported, yet.")

        return maps

    def get_projector_args(
        self,
        roi: Optional[ArrayLike] = None,
        rot_ind: Union[int, slice, Sequence[int], NDArrayInt, None] = None,
        det_ind: Union[int, slice, Sequence[int], NDArrayInt, None] = None,
    ) -> dict[str, NDArray]:
        """
        Return the projector arguments.

        Parameters
        ----------
        roi : ArrayLike, optional
            The region-of-interest to select. The default is None.
        rot_ind : int, optional
            A specific rotation index, if only one is to be desired. The default is None.
        det_ind : int, optional
            A specific detector index, if only one is to be desired. The default is None.

        Returns
        -------
        dict[str, NDArray]
            A dictionary containing the attenuation maps and the detector angle.
        """
        if det_ind is None:
            det_angles = self.angles_det_rad
        else:
            det_angles = self.angles_det_rad[det_ind]
        return dict(att_maps=self.get_maps(roi=roi, rot_ind=rot_ind, det_ind=det_ind), angles_detectors_rad=det_angles)


def get_linear_attenuation_coefficient(
    compound: Union[str, dict], energy_keV: float, pixel_size_um: float, density: Union[float, None] = None
) -> float:
    """Compute the linear attenuation coefficient for given compound, energy, and pixel size.

    Parameters
    ----------
    compound : Union[str, dict]
        The compound for which we compute the linear attenuation coefficient
    energy_keV : float
        The energy of the photons
    pixel_size_um : float
        The pixel size in microns
    density : Union[float, None], optional
        The density of the compound (if different from the default value), by default None

    Returns
    -------
    float
        The linear attenuation coefficient
    """
    if isinstance(compound, str):
        compound = get_compound(compound)

    if density is not None:
        compound["density"] = density

    cmp_cs = get_compound_cross_section(compound, energy_keV)
    return pixel_size_um * CONVERT_UM_TO_CM * compound["density"] * cmp_cs


def plot_emission_line_attenuation(
    compound: Union[str, dict],
    thickness_um: float,
    mean_energy_keV: float,
    fwhm_keV: float,
    line_shape: str = "lorentzian",
    num_points: int = 201,
    plot_lines_mean: bool = True,
) -> None:
    """Plot spectral attenuation of a given line.

    Parameters
    ----------
    compound : Union[str, dict]
        Compound to consider
    thickness_um : float
        Thickness of the compound (in microns)
    mean_energy_keV : float
        Average energy of the line
    fwhm_keV : float
        Full-width half-maximum of the line
    line_shape : str, optional
        Shape of the line, by default "lorentzian".
        Options are: "gaussian" | "lorentzian" | "sech**2".
    num_points : int, optional
        number of discretization points, by default 201
    plot_lines_mean : bool, optional
        Whether to plot the line mean and the effective line mean, by default True

    Raises
    ------
    ValueError
        When an unsupported line is chosen.
    """
    xc = np.linspace(-0.5, 0.5, num_points)

    if line_shape.lower() == "gaussian":
        xc *= fwhm_keV * 3
        yg = np.exp(-4 * np.log(2) * (xc**2) / (fwhm_keV**2))
    elif line_shape.lower() == "lorentzian":
        xc *= fwhm_keV * 13
        hwhm_keV = fwhm_keV / 2
        yg = hwhm_keV / (xc**2 + hwhm_keV**2)
    elif line_shape.lower() == "sech**2":
        # doi: 10.1364/ol.20.001160
        xc *= fwhm_keV * 4
        tau = fwhm_keV / (2 * np.arccosh(np.sqrt(2)))
        yg = 1 / np.cosh(xc / tau) ** 2
    else:
        raise ValueError(f"Unknown beam shape: {line_shape.lower()}")

    nrgs_keV = xc + mean_energy_keV

    if isinstance(compound, str):
        compound = get_compound(compound)

    atts = np.empty_like(yg)

    for ii, nrg in enumerate(nrgs_keV):
        cmp_cs = get_compound_cross_section(compound, nrg)
        atts[ii] = np.exp(-thickness_um * CONVERT_UM_TO_CM * compound["density"] * cmp_cs)

    yg = yg / np.max(yg)
    mean_effective_energy_keV = float(nrgs_keV.dot(yg * atts) / yg.dot(atts))

    fig, axs_line = plt.subplots(1, 1)
    pl_line = axs_line.plot(nrgs_keV, yg, label="$I_0$", color="C0")
    if plot_lines_mean:
        axs_line.axvline(mean_energy_keV, linestyle="--", color="C0")
    axs_line.tick_params(axis="y", labelcolor="C0", labelsize=14)
    axs_line.tick_params(axis="x", labelsize=14)
    axs_line.set_ylabel("Intensity", color="C0", fontsize=14)
    axs_atts = axs_line.twinx()
    pl_atts = axs_atts.plot(nrgs_keV, atts, label="exp(-$\\mu (E) x)$", color="C1")
    pl_line_att = axs_atts.plot(nrgs_keV, yg * atts, label=r"$I_{measured}$", color="C2")
    if plot_lines_mean:
        axs_atts.axvline(mean_effective_energy_keV, linestyle="--", color="C2")
    axs_atts.tick_params(axis="y", labelcolor="C1", labelsize=14)
    axs_atts.set_ylabel("Transmittance", color="C1", fontsize=14)
    all_pls = pl_line + pl_atts + pl_line_att
    axs_atts.legend(all_pls, [str(pl.get_label()) for pl in all_pls], fontsize=13)
    axs_line.grid()
    fig.tight_layout()

    I_lin = np.sum(yg * atts[num_points // 2])
    I_meas = yg.dot(atts)
    print(f"Expected intensity: {I_lin}, measured: {I_meas} ({I_meas / I_lin:%})")
    print(f"Expected mean energy: {nrgs_keV.dot(yg / np.sum(yg)):.5}, effective: {mean_effective_energy_keV:.5}")


def plot_transmittance_decay(
    compounds: Union[str, dict, Sequence[Union[str, dict]]],
    mean_energy_keV: float,
    thickness_range_um: tuple[float, float, int] = (0.0, 10.0, 101),
) -> None:
    """Plot transmittance decay curve(s) for the given compound(s) at a given energy and thickness range.

    Parameters
    ----------
    compounds : str | dict | Sequence[str | dict]
        The compound(s) description
    mean_energy_keV : float
        The mean photon energy
    thickness_range_um : tuple[float, float, int], optional
        The thickness range as (start, end, num_points), by default (0.0, 10.0, 101)
    """
    if isinstance(compounds, (str, dict)):
        compounds = [compounds]

    compounds = [get_compound(c) if isinstance(c, str) else c for c in compounds]

    thicknesses_um = np.linspace(*thickness_range_um)

    atts = np.zeros((len(compounds), len(thicknesses_um)), dtype=np.float32)
    for ii, cmp in enumerate(compounds):
        cmp_cs = get_compound_cross_section(cmp, mean_energy_keV)
        atts[ii] = np.exp(-thicknesses_um * CONVERT_UM_TO_CM * cmp["density"] * cmp_cs)

    fig, axs = plt.subplots(1, 1, figsize=(8, 4))
    for ii, cmp in enumerate(compounds):
        axs.plot(thicknesses_um, atts[ii], label=cmp["name"])
    axs.legend(fontsize=13)
    axs.grid()
    axs.tick_params(labelsize=14)
    axs.set_xlabel("Thickness [$\mu m$]", fontsize=14)
    axs.set_xlim(thickness_range_um[0], thickness_range_um[1])
    axs.set_ylabel("Transmittance", fontsize=14)
    axs.set_ylim(0.0, 1.0)
    axs.set_title(f"Transmittance curve at {mean_energy_keV:.2f} keV", fontsize=14)
    fig.tight_layout()
    plt.plot(block=False)



#!/usr/bin/env python3
"""
Materials support functions and classes.

@author: Nicola VIGANÒ, ESRF - The European Synchrotron, Grenoble, France,
and CEA-IRIG, Grenoble, France
"""

import copy as cp
from collections.abc import Sequence
from typing import Union

import numpy as np
from numpy.typing import DTypeLike, NDArray

from corrct.physics import xrf  # noqa: F401, F402
from corrct.physics.xraylib_helper import get_compound_cross_section
from corrct.physics.xraylib_helper import get_element_number
from corrct.physics.xraylib_helper import xraylib


class VolumeMaterial:
    """
    VolumeMaterial class, that can be used for predicting fluorescence and Compton productions, attenuation, etc.

    Parameters
    ----------
    material_fractions : Sequence
        Concentration fractions of each material for each voxel.
    material_compounds : Sequence
        Compound description of each material.
    voxel_size_cm : float
        Voxel size in cm.
    dtype : DTypeLike, optional
        Data type of the produced data. The default is None.

    Raises
    ------
    ValueError
        Raised in case of incorrect parameters.
    """

    def __init__(
        self,
        materials_fractions: Sequence[NDArray],
        materials_composition: Sequence,
        voxel_size_cm: float,
        dtype: DTypeLike = None,
        verbose: bool = False,
    ):
        if len(materials_fractions) != len(materials_composition):
            raise ValueError(
                f"Materials fractions (# {len(materials_fractions)}) and "
                f"materials composition (# {len(materials_composition)}) arrays should have the same length"
            )
        if len(materials_fractions) == 0:
            raise ValueError("Phase list is empty")

        self.materials_fractions = list(materials_fractions)
        self.shape = np.array(self.materials_fractions[0].shape)
        for ii, ph in enumerate(self.materials_fractions):
            if len(self.shape) != len(ph.shape):
                raise ValueError("All material fraction volumes should have the same number of dimensions")
            if not np.all(self.shape == ph.shape):
                raise ValueError("Materials fraction volumes should all have the same shape")
            if ph.dtype == bool:
                self.materials_fractions[ii] = ph.astype(np.float32)

        if dtype is None:
            dtype = self.materials_fractions[0].dtype
        self.dtype = dtype
        for ii, ph in enumerate(self.materials_fractions):
            self.materials_fractions[ii] = ph.astype(self.dtype)

        self.materials_compositions = [
            xraylib.GetCompoundDataNISTByName(cmp) if isinstance(cmp, str) else cmp for cmp in materials_composition
        ]

        self.voxel_size_cm = voxel_size_cm
        self.verbose = verbose

    def get_attenuation(self, energy_keV: float) -> NDArray:
        """
        Compute the local attenuation for each voxel.

        Parameters
        ----------
        energy_keV : float
            The X-ray energy.

        Returns
        -------
        NDArray
            The computed local attenuation volume.
        """
        ph_lin_att = np.zeros(self.shape, self.dtype)
        for ph, cmp in zip(self.materials_fractions, self.materials_compositions):
            cmp_cs = get_compound_cross_section(cmp, energy_keV)
            if self.verbose:
                print(f"Attenuation ({cmp['name']} at {energy_keV}):")
                print(
                    f" - cross-section * mass fraction = {cmp_cs}, density = {cmp['density']}, pixel-size {self.voxel_size_cm}"
                )
                print(f" - total {cmp['density'] * cmp_cs * self.voxel_size_cm} (assuming material mass fraction = 1)")
            ph_lin_att += ph * cmp["density"] * cmp_cs
        return ph_lin_att * self.voxel_size_cm

    def get_element_mass_fraction(self, element: Union[str, int]) -> NDArray:
        """Compute the local element mass fraction through out all the materials.

        Parameters
        ----------
        element : str | int
            The element to look for.

        Returns
        -------
        mass_fraction : NDArray
            The local mass fraction in each voxel.
        """
        el_num = get_element_number(element)

        mass_fraction = np.zeros(self.shape, self.dtype)
        for ph, cmp in zip(self.materials_fractions, self.materials_compositions):
            if el_num in cmp["Elements"]:
                el_ind = np.where(np.array(cmp["Elements"]) == el_num)[0][0]
                mass_fraction += ph * cmp["density"] * cmp["massFractions"][el_ind]
        return mass_fraction

    def _check_parallax_detector(self, detector: xrf.DetectorXRF, tolerance: float = 1e-2) -> bool:
        half_sample_size = np.max(self.voxel_size_cm * self.shape) / 2

        if isinstance(detector.distance_mm, float) or (
            isinstance(detector.distance_mm, np.ndarray) and detector.distance_mm.size == 1
        ):
            dets = cp.deepcopy(detector)
            dets.distance_mm = dets.distance_mm + np.array([-half_sample_size, half_sample_size])
        else:
            dets = detector

        solid_angles = dets.solid_angle_sr
        max_parallax = np.max(solid_angles) - np.min(solid_angles)
        return max_parallax < tolerance

    def get_compton_scattering(
        self, energy_in_keV: float, angle_rad: Union[float, None] = None, detector: Union[xrf.DetectorXRF, None] = None
    ) -> tuple[float, NDArray]:
        """Compute the local Compton scattering.

        Parameters
        ----------
        energy_in_keV : float
            Incoming beam energy.
        angle_rad : float, optional
            The detector angle, with respect to incoming beam direction. The default is None.
        detector : DetectorXRF, optional
            The detector object. The default is None.

        Raises
        ------
        ValueError
            In case neither of `angle_rad` nor `detector` have been passed.

        Returns
        -------
        energy_out_keV : float
            The energy of the Compton radiation received by the detector.
        cmptn_prod : NDArray
            Local production of Compton radiation.

        Either `angle_rad` or `detector` need to be supplied.
        """
        if detector is None:
            if angle_rad is None:
                raise ValueError("Either 'angle_rad' or 'detector' should be passed.")
        else:
            if not self._check_parallax_detector(detector):
                print("WARNING - detector parallax is above 1e-2")
            if angle_rad is None:
                angle_rad = detector.angle_rad

        cmptn_prod = np.zeros(self.shape, self.dtype)
        for ph, cmp in zip(self.materials_fractions, self.materials_compositions):
            try:
                cmp_cs = xraylib.DCS_Compt_CP(cmp["name"], energy_in_keV, angle_rad)
            except ValueError:
                cmp_cs = np.sum(
                    [
                        xraylib.DCS_Compt(el, energy_in_keV, angle_rad) * cmp["massFractions"][ii]
                        for ii, el in enumerate(cmp["Elements"])
                    ],
                    axis=0,
                )
            if self.verbose:
                print(
                    f"Compton - {cmp['name']} at incoming energy {energy_in_keV} (keV),"
                    + f" outgoing angle {np.rad2deg(angle_rad)} (deg):\n"
                    + f" - cross-section * mass fraction = {cmp_cs}, density = {cmp['density']}"
                    + ", pixel-size {self.voxel_size_cm}"
                    + f" - total {cmp['density'] * cmp_cs * self.voxel_size_cm} (assuming material mass fraction = 1)"
                )
            cmptn_prod += ph * cmp["density"] * cmp_cs
        cmptn_prod *= self.voxel_size_cm

        if detector:
            cmptn_prod *= detector.solid_angle_sr

        energy_out_keV = xraylib.ComptonEnergy(energy_in_keV, angle_rad)
        return (energy_out_keV, cmptn_prod)

    def get_fluo_production(
        self,
        element: Union[str, int],
        energy_in_keV: float,
        fluo_lines: Union[str, xrf.FluoLine, Sequence[xrf.FluoLine]],
        detector: Union[xrf.DetectorXRF, None] = None,
    ) -> tuple[float, NDArray]:
        """Compute the local fluorescence production, for the given line of the given element.

        Using Eq. (1) from:
        - T. Schoonjans et al., "The xraylib library for X-ray-matter interactions. Recent developments," Spectrochim. Acta
        Part B At. Spectrosc., vol. 66, no. 11-12, pp. 776-784, Nov. 2011, doi: 10.1016/j.sab.2011.09.011.

        Parameters
        ----------
        element : str | int
            The element to consider.
        energy_in_keV : float
            The incoming X-ray beam energy.
        fluo_lines : str | FluoLine | Sequence[FluoLine]
            The fluorescence line to consider.
        detector : DetectorXRF, optional
            The detector geometry. The default is None.

        Returns
        -------
        energy_out_keV : float
            The emitted fluorescence energy.
        el_prod : NDArray
            The local fluorescence production in each voxel.
        """
        if detector:
            if not self._check_parallax_detector(detector):
                print("WARNING - detector parallax is above 1e-2")

        if isinstance(fluo_lines, xrf.FluoLine):
            fluo_lines = [fluo_lines]
        elif isinstance(fluo_lines, str):
            fluo_lines = xrf.LinesSiegbahn.get_lines(fluo_lines)

        el_num = get_element_number(element)

        el_yield = np.empty((len(fluo_lines),), self.dtype)
        for ii, line in enumerate(fluo_lines):
            try:
                el_yield[ii] = xraylib.CS_FluorLine_Kissel(el_num, line.indx, energy_in_keV)  # fluo production for cm2/g
            except ValueError as exc:
                el_sym = xraylib.AtomicNumberToSymbol(el_num)
                if self.verbose:
                    print(f"Energy {exc}: el_num={el_num} ({el_sym}) line={line}")
                el_yield[ii] = 0
        el_prod = self.get_element_mass_fraction(el_num) * np.sum(el_yield) * self.voxel_size_cm

        if detector:
            el_prod *= detector.solid_angle_sr

        energy_out_keV = xrf.get_energy(el_num, fluo_lines, compute_average=True)

        return float(energy_out_keV), el_prod



#!/usr/bin/env python3
"""
XRF support functions.

@author: Nicola VIGANÒ, ESRF - The European Synchrotron, Grenoble, France,
and CEA-IRIG, Grenoble, France
"""

try:
    from . import xraylib_helper  # noqa: F401, F402

    xraylib = xraylib_helper.xraylib

except ImportError:
    print("WARNING: Physics support is only available when xraylib is installed!")
    raise


from collections.abc import Sequence
from dataclasses import dataclass
from typing import overload, Literal, Union
import numpy as np
from numpy.typing import NDArray


@dataclass
class FluoLine:
    """Fluorescence line description class."""

    name: str
    indx: int


class LinesSiegbahn:
    """Siegbahn fluorescence lines collection class."""

    lines = [
        FluoLine(name="KA1", indx=xraylib.KA1_LINE),
        FluoLine(name="KA2", indx=xraylib.KA2_LINE),
        FluoLine(name="KA3", indx=xraylib.KA3_LINE),
        FluoLine(name="KB1", indx=xraylib.KB1_LINE),
        FluoLine(name="KB2", indx=xraylib.KB2_LINE),
        FluoLine(name="KB3", indx=xraylib.KB3_LINE),
        FluoLine(name="KB4", indx=xraylib.KB4_LINE),
        FluoLine(name="KB5", indx=xraylib.KB5_LINE),
        FluoLine(name="LA1", indx=xraylib.LA1_LINE),
        FluoLine(name="LA2", indx=xraylib.LA2_LINE),
        FluoLine(name="LB1", indx=xraylib.LB1_LINE),
        FluoLine(name="LB2", indx=xraylib.LB2_LINE),
        FluoLine(name="LB3", indx=xraylib.LB3_LINE),
        FluoLine(name="LB4", indx=xraylib.LB4_LINE),
        FluoLine(name="LB5", indx=xraylib.LB5_LINE),
        FluoLine(name="LB6", indx=xraylib.LB6_LINE),
        FluoLine(name="LB7", indx=xraylib.LB7_LINE),
        FluoLine(name="LB9", indx=xraylib.LB9_LINE),
        FluoLine(name="LB10", indx=xraylib.LB10_LINE),
        FluoLine(name="LB15", indx=xraylib.LB15_LINE),
        FluoLine(name="LB17", indx=xraylib.LB17_LINE),
        FluoLine(name="LG1", indx=xraylib.LG1_LINE),
        FluoLine(name="LG2", indx=xraylib.LG2_LINE),
        FluoLine(name="LG3", indx=xraylib.LG3_LINE),
        FluoLine(name="LG4", indx=xraylib.LG4_LINE),
        FluoLine(name="LG5", indx=xraylib.LG5_LINE),
        FluoLine(name="LG6", indx=xraylib.LG6_LINE),
        FluoLine(name="LG8", indx=xraylib.LG8_LINE),
        FluoLine(name="LE", indx=xraylib.LE_LINE),
        FluoLine(name="LH", indx=xraylib.LH_LINE),
        FluoLine(name="LL", indx=xraylib.LL_LINE),
        FluoLine(name="LS", indx=xraylib.LS_LINE),
        FluoLine(name="LT", indx=xraylib.LT_LINE),
        FluoLine(name="LU", indx=xraylib.LU_LINE),
        FluoLine(name="LV", indx=xraylib.LV_LINE),
        FluoLine(name="MA1", indx=xraylib.MA1_LINE),
        FluoLine(name="MA2", indx=xraylib.MA2_LINE),
        FluoLine(name="MB", indx=xraylib.MB_LINE),
        FluoLine(name="MG", indx=xraylib.MG_LINE),
    ]

    @staticmethod
    def get_lines(line: str) -> Sequence[FluoLine]:
        """
        Return the list of xraylib line macro definitions for the requested family.

        Parameters
        ----------
        line : str
            The requested line. It can be a whole shell (transition to that shell),
            or sub-shells.

        Returns
        -------
        Sequence
            List of corresponding lines.
        """
        return [f for f in LinesSiegbahn.lines if f.name[: len(line)] == line.upper()]


def _get_lines_list(lines) -> Sequence[FluoLine]:
    if isinstance(lines, FluoLine):
        return [lines]
    elif isinstance(lines, str):
        return LinesSiegbahn.get_lines(lines)
    elif len(lines) == 0:
        raise ValueError(f"No line was passed! lines={lines}")
    else:
        return lines


def get_radiation_rate(
    element: Union[str, int],
    lines: Union[str, FluoLine, Sequence[FluoLine]],
    verbose: bool = False,
) -> NDArray:
    """Return the radiation rates of the requested lines for the given element.

    Parameters
    ----------
    element : Union[str, int]
        The requested element
    lines : Union[str, FluoLine, Sequence[FluoLine]]
        The requested line. It can be a whole shell (transition to that shell),
        or sub-shells.
    verbose : bool, optional
        Whether to produce verbose output in case of errors, by default False

    Returns
    -------
    NDArray
        The list of radiation rates
    """
    el_sym, el_num = xraylib_helper.get_element_number_and_symbol(element)

    lines_list = _get_lines_list(lines)

    rates = np.empty(len(lines_list), dtype=np.float32)
    for ii, line in enumerate(lines_list):
        try:
            rates[ii] = xraylib.RadRate(el_num, line.indx)
        except ValueError as exc:
            if verbose:
                print(f"INFO - RadRate - {exc}: el_num={el_num} ({el_sym}) line={line}")
            rates[ii] = 0
    return rates


@overload
def get_energy(
    element: Union[str, int],
    lines: Union[str, FluoLine, Sequence[FluoLine]],
    *,
    compute_average: Literal[False] = False,
    verbose: bool = False,
) -> NDArray: ...


@overload
def get_energy(
    element: Union[str, int],
    lines: Union[str, FluoLine, Sequence[FluoLine]],
    *,
    compute_average: Literal[True] = True,
    verbose: bool = False,
) -> float: ...


def get_energy(
    element: Union[str, int],
    lines: Union[str, FluoLine, Sequence[FluoLine]],
    *,
    compute_average: bool = False,
    verbose: bool = False,
) -> Union[float, NDArray]:
    """
    Return the energy(ies) of the requested line for the given element.

    Parameters
    ----------
    element : Union[str, int]
        The requested element.
    line : str
        The requested line. It can be a whole shell (transition to that shell),
        or sub-shells.
    compute_average : bool, optional
        Weighted averaging the lines, using the radiation rate. The default is False.

    Returns
    -------
    energy_keV : Union[float, NDArray]
        Either the average energy or the list of different energies.
    """
    el_sym, el_num = xraylib_helper.get_element_number_and_symbol(element)

    lines_list = _get_lines_list(lines)

    energy_keV = np.empty(len(lines_list), dtype=np.float32)
    for ii, line in enumerate(lines_list):
        try:
            energy_keV[ii] = xraylib.LineEnergy(el_num, line.indx)
        except ValueError as exc:
            if verbose:
                print(f"INFO - Energy - {exc}: el_num={el_num} ({el_sym}) line={line}")
            energy_keV[ii] = 0

    if compute_average:
        rates = get_radiation_rate(element, lines_list)
        energy_keV = float(np.sum(energy_keV * rates / np.sum(rates)))

    if verbose:
        print(f"{el_sym}-{lines} emission energy (keV):", energy_keV, "\n")

    return energy_keV


@dataclass
class DetectorXRF:
    """Simple XRF detector model."""

    surface_mm2: float
    distance_mm: Union[float, NDArray]
    angle_rad: float = np.pi / 2

    @property
    def solid_angle_sr(self) -> Union[float, NDArray]:
        """Compute the solid angle covered by the detector.

        Returns
        -------
        float | NDArray
            The computed solid angle of the detector geometry.
        """
        return self.surface_mm2 / (4 * np.pi * self.distance_mm**2)



#!/usr/bin/env python3
"""
Solvers for the tomographic reconstruction problem.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import copy as cp
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Union
from collections.abc import Sequence

import numpy as np
from numpy.typing import DTypeLike, NDArray
from tqdm.auto import tqdm

from . import data_terms, filters, operators, projectors, regularizers

eps = np.finfo(np.float32).eps


NDArrayFloat = NDArray[np.floating]


class SolutionInfo:
    """Reconstruction info."""

    method: str
    iterations: int
    max_iterations: int

    residual0: Union[float, np.floating]
    residual0_cv: Union[float, np.floating]

    residuals: NDArrayFloat
    residuals_cv: NDArrayFloat
    tolerance: Union[float, np.floating, None]

    def __init__(
        self,
        method: str,
        max_iterations: int,
        tolerance: Union[float, np.floating, None],
        residual0: float = np.inf,
        residual0_cv: float = np.inf,
    ) -> None:
        self.method = method
        self.max_iterations = max_iterations
        self.tolerance = tolerance

        self.residual0 = residual0
        self.residual0_cv = residual0_cv

        self.residuals = np.zeros(max_iterations)
        self.residuals_cv = np.zeros(max_iterations)

        self.iterations = 0

    @property
    def residuals_rel(self) -> NDArrayFloat:
        return self.residuals / self.residual0

    @property
    def residuals_cv_rel(self) -> NDArrayFloat:
        return self.residuals_cv / self.residual0_cv


class Solver(ABC):
    """
    Initialize the base solver class.

    Parameters
    ----------
    verbose : bool, optional
        Turn on verbose output. The default is False.
    tolerance : Optional[float], optional
        Tolerance on the data residual for computing when to stop iterations.
        The default is None.
    relaxation : float, optional
        The relaxation length. The default is 1.0.
    data_term : Union[str, data_terms.DataFidelityBase], optional
        Data fidelity term for computing the data residual. The default is "l2".
    data_term_test : Optional[data_terms.DataFidelityBase], optional
        The data fidelity to be used for the test set.
        If None, it will use the same as for the rest of the data.
        The default is None.
    """

    def __init__(
        self,
        verbose: bool = False,
        leave_progress: bool = True,
        relaxation: float = 1.0,
        tolerance: Optional[float] = None,
        data_term: Union[str, data_terms.DataFidelityBase] = "l2",
        data_term_test: Union[str, data_terms.DataFidelityBase, None] = None,
    ):
        self.verbose = verbose
        self.leave_progress = leave_progress
        self.relaxation = relaxation
        self.tolerance = tolerance

        self.data_term = self._initialize_data_fidelity_function(data_term)
        if data_term_test is None:
            data_term_test = self.data_term
        else:
            data_term_test = self._initialize_data_fidelity_function(data_term_test)
        self.data_term_test = cp.deepcopy(data_term_test)

    def info(self) -> str:
        """
        Return the solver info.

        Returns
        -------
        str
            Solver info string.
        """
        return type(self).__name__

    def upper(self) -> str:
        """
        Return the upper case name of the solver.

        Returns
        -------
        str
            Upper case string name of the solver.
        """
        return type(self).__name__.upper()

    def lower(self) -> str:
        """
        Return the lower case name of the solver.

        Returns
        -------
        str
            Lower case string name of the solver.
        """
        return type(self).__name__.lower()

    @abstractmethod
    def __call__(
        self, A: operators.BaseTransform, b: NDArrayFloat, *args: Any, **kwds: Any
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """Execute the reconstruction of the data.

        Parameters
        ----------
        A : operators.BaseTransform
            The projection operator.
        b : NDArrayFloat
            The data to be reconstructed.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction and related information.
        """

    @staticmethod
    def _initialize_data_fidelity_function(data_term: Union[str, data_terms.DataFidelityBase]) -> data_terms.DataFidelityBase:
        if isinstance(data_term, str):
            if data_term.lower() == "l2":
                return data_terms.DataFidelity_l2()
            elif data_term.lower() == "kl":
                return data_terms.DataFidelity_KL()
            else:
                raise ValueError(f"Unknown data term: '{data_term}', only accepted terms are: 'l2' | 'kl'.")
        elif isinstance(data_term, (data_terms.DataFidelity_l2, data_terms.DataFidelity_KL)):
            return cp.deepcopy(data_term)
        else:
            raise ValueError(f"Unsupported data term: '{data_term.info()}', only accepted terms are 'kl' and 'l2'-based.")

    @staticmethod
    def _initialize_regularizer(
        regularizer: Union[regularizers.BaseRegularizer, None, Sequence[regularizers.BaseRegularizer]]
    ) -> Sequence[regularizers.BaseRegularizer]:
        if regularizer is None:
            return []
        elif isinstance(regularizer, regularizers.BaseRegularizer):
            return [regularizer]
        elif isinstance(regularizer, (list, tuple)):
            check_regs_ok = [isinstance(r, regularizers.BaseRegularizer) for r in regularizer]
            if not np.all(check_regs_ok):
                raise ValueError(
                    "The following regularizers are not derived from the regularizers.BaseRegularizer class: "
                    f"{np.array(np.arange(len(check_regs_ok))[np.array(check_regs_ok, dtype=bool)])}"
                )
            else:
                return list(regularizer)
        else:
            raise ValueError("Unknown regularizer type.")

    @staticmethod
    def _initialize_b_masks(
        b: NDArrayFloat, b_mask: Optional[NDArrayFloat], b_test_mask: Optional[NDArrayFloat]
    ) -> tuple[Optional[NDArrayFloat], Optional[NDArrayFloat]]:
        if b_test_mask is not None:
            if b_mask is None:
                b_mask = np.ones_like(b)
            # As we are being passed a test residual pixel mask, we need
            # to make sure to mask those pixels out from the reconstruction.
            # At the same time, we need to remove any masked pixel from the test count.
            b_mask, b_test_mask = b_mask * (1 - b_test_mask), b_test_mask * b_mask
        return (b_mask, b_test_mask)


class FBP(Solver):
    """Implementation of the Filtered Back-Projection (FBP) algorithm."""

    def __init__(
        self,
        verbose: bool = False,
        leave_progress: bool = False,
        regularizer: Union[Sequence[regularizers.BaseRegularizer], regularizers.BaseRegularizer, None] = None,
        data_term: Union[str, data_terms.DataFidelityBase] = "l2",
        fbp_filter: Union[str, NDArrayFloat, filters.Filter] = "ramp",
        pad_mode: str = "constant",
    ):
        """Initialize the Filtered Back-Projection (FBP) algorithm.

        Parameters
        ----------
        verbose : bool, optional
            Turn on verbose output. The default is False.
        leave_progress: bool, optional
            Leave the progress bar after the computation is finished. The default is True.
        regularizer : Sequence[regularizers.BaseRegularizer] | regularizers.BaseRegularizer | None, optional
            NOT USED, only exposed for compatibility reasons.
        data_term : Union[str, data_terms.DataFidelityBase], optional
            NOT USED, only exposed for compatibility reasons.
        fbp_filter : Union[str, NDArrayFloat], optional
            FBP filter to use. Either a string from scikit-image's list of `iradon` filters, or an array. The default is "ramp".
        pad_mode: str, optional
            The padding mode to use for the linear convolution. The default is "constant".
        """
        super().__init__(verbose=verbose)
        if isinstance(fbp_filter, str):
            fbp_filter = fbp_filter.lower()
        self.fbp_filter = fbp_filter
        self.pad_mode = pad_mode

    def info(self) -> str:
        """
        Return the solver info.

        Returns
        -------
        str
            Solver info string.
        """
        if isinstance(self.fbp_filter, str):
            return super().info() + "(F:" + self.fbp_filter.upper() + ")"
        elif isinstance(self.fbp_filter, np.ndarray):
            return super().info() + "(F:" + filters.FilterCustom.__name__.upper() + ")"
        else:
            return super().info() + "(F:" + type(self.fbp_filter).__name__.upper() + ")"

    def __call__(  # noqa: C901
        self,
        A: operators.BaseTransform,
        b: NDArrayFloat,
        iterations: int = 0,
        x0: Optional[NDArrayFloat] = None,
        lower_limit: Union[float, NDArrayFloat, None] = None,
        upper_limit: Union[float, NDArrayFloat, None] = None,
        x_mask: Optional[NDArrayFloat] = None,
        b_mask: Optional[NDArrayFloat] = None,
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """
        Reconstruct the data, using the FBP algorithm.

        Parameters
        ----------
        A : BaseTransform
            Projection operator.
        b : NDArrayFloat
            Data to reconstruct.
        iterations : int
            Number of iterations.
        x0 : Optional[NDArrayFloat], optional
            Initial solution. The default is None.
        lower_limit : Union[float, NDArrayFloat], optional
            Lower clipping value. The default is None.
        upper_limit : Union[float, NDArrayFloat], optional
            Upper clipping value. The default is None.
        x_mask : Optional[NDArrayFloat], optional
            Solution mask. The default is None.
        b_mask : Optional[NDArrayFloat], optional
            Data mask. The default is None.

        Raises
        ------
        ValueError
            In case the data is 1D.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction, and None.
        """
        if len(b.shape) < 2:
            raise ValueError(f"Data should be at least 2-dimensional (b.shape = {b.shape})")

        info = SolutionInfo(self.info(), max_iterations=0, tolerance=0.0)

        if isinstance(self.fbp_filter, str):
            if self.fbp_filter in ("mr", "data"):
                local_filter = filters.FilterMR(projector=A)
            else:
                local_filter = filters.FilterFBP(filter_name=self.fbp_filter)
        elif isinstance(self.fbp_filter, np.ndarray):
            local_filter = filters.FilterCustom(self.fbp_filter)
        else:
            local_filter = self.fbp_filter
        local_filter.pad_mode = self.pad_mode

        if isinstance(A, operators.ProjectorOperator):
            pre_weights = A.get_pre_weights()
            if pre_weights is not None:
                b = b * pre_weights

        b_f = local_filter(b)

        x = A.T(b_f)

        if lower_limit is not None or upper_limit is not None:
            x = x.clip(lower_limit, upper_limit)
        if x_mask is not None:
            x *= x_mask

        return x, info


class SART(Solver):
    """Solver class implementing the Simultaneous Algebraic Reconstruction Technique (SART) algorithm."""

    def compute_residual(
        self,
        A: Callable,
        b: NDArrayFloat,
        x: NDArrayFloat,
        A_num_rows: int,
        b_mask: Optional[NDArrayFloat],
    ) -> NDArrayFloat:
        """Compute the solution residual.

        Parameters
        ----------
        A : Callable
            The forward projector.
        b : NDArrayFloat
            The detector data.
        x : NDArrayFloat
            The current solution
        A_num_rows : int
            The number of projections.
        b_mask : Optional[NDArrayFloat]
            The mask to apply

        Returns
        -------
        NDArrayFloat
            The residual.
        """
        fp = np.stack([A(x, ii) for ii in range(A_num_rows)], axis=-1)
        fp = np.ascontiguousarray(fp, dtype=b.dtype)
        res = fp - b
        if b_mask is not None:
            res *= b_mask
        return res

    def __call__(  # noqa: C901
        self,
        A: Union[Callable[[NDArray, int], NDArray], projectors.ProjectorUncorrected],
        b: NDArrayFloat,
        iterations: int,
        A_num_rows: Optional[int] = None,
        At: Optional[Callable] = None,
        x0: Optional[NDArrayFloat] = None,
        lower_limit: Union[float, NDArrayFloat, None] = None,
        upper_limit: Union[float, NDArrayFloat, None] = None,
        x_mask: Optional[NDArrayFloat] = None,
        b_mask: Optional[NDArrayFloat] = None,
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """
        Reconstruct the data, using the SART algorithm.

        Parameters
        ----------
        A : Union[Callable, BaseTransform]
            Projection operator.
        b : NDArrayFloat
            Data to reconstruct.
        iterations : int
            Number of iterations.
        A_num_rows : int
            Number of projections.
        x0 : Optional[NDArrayFloat], optional
            Initial solution. The default is None.
        At : Callable, optional
            The back-projection operator. This is only needed if the projection operator does not have an adjoint.
            The default is None.
        lower_limit : Union[float, NDArrayFloat], optional
            Lower clipping value. The default is None.
        upper_limit : Union[float, NDArrayFloat], optional
            Upper clipping value. The default is None.
        x_mask : Optional[NDArrayFloat], optional
            Solution mask. The default is None.
        b_mask : Optional[NDArrayFloat], optional
            Data mask. The default is None.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction, and the residuals.
        """
        if isinstance(A, projectors.ProjectorUncorrected):
            p = A

            if not p.projector_backend.has_individual_projs:
                raise ValueError("The projector needs to have enabled single projections.")

            A = lambda x, ii: p.fp_angle(x, ii)  # noqa: E731
            if isinstance(p, projectors.ProjectorAttenuationXRF):
                At = lambda y, ii: p.bp_angle(y, ii, single_line=True)  # noqa: E731
            else:
                At = lambda y, ii: p.bp_angle(y, ii)  # noqa: E731
            A_num_rows = len(p.angles_rot_rad)
        elif At is None:
            raise ValueError("Parameter `At` is required, if `A` is not a projector.")
        elif A_num_rows is None:
            raise ValueError("Parameter `A_num_rows` is required, if `A` is not a projector.")

        # Back-projection diagonal re-scaling
        b_ones = np.ones_like(b)
        if b_mask is not None:
            b_ones *= b_mask
        tau = [At(b_ones[..., ii, :], ii) for ii in range(A_num_rows)]
        tau = np.abs(np.stack(tau, axis=-2))
        tau[(tau / np.max(tau)) < 1e-5] = 1
        tau = self.relaxation / tau

        # Forward-projection diagonal re-scaling
        x_ones = np.ones([*tau.shape[:-2], tau.shape[-1]], dtype=tau.dtype)
        if x_mask is not None:
            x_ones *= x_mask
        sigma = [A(x_ones, ii) for ii in range(A_num_rows)]
        sigma = np.abs(np.stack(sigma, axis=-2))
        sigma[(sigma / np.max(sigma)) < 1e-5] = 1
        sigma = 1 / sigma

        if x0 is None:
            x0 = np.zeros_like(x_ones)
        else:
            x0 = np.array(x0).copy()
        x = x0

        info = SolutionInfo(self.info(), max_iterations=iterations, tolerance=self.tolerance)

        if self.tolerance is not None:
            res = self.compute_residual(A, b, x, A_num_rows=A_num_rows, b_mask=b_mask)
            info.residual0 = np.linalg.norm(res.flatten())

        rows_sequence = np.random.permutation(A_num_rows)

        algo_info = f"- Performing {self.upper()} iterations: "

        for ii in tqdm(range(iterations), desc=algo_info, disable=(not self.verbose), leave=self.leave_progress):
            info.iterations += 1

            for ii_a in rows_sequence:
                res = A(x, ii_a) - b[..., ii_a, :]
                if b_mask is not None:
                    res *= b_mask[..., ii_a, :]

                x -= At(res * sigma[..., ii_a, :], ii_a) * tau[..., ii_a, :]

                if lower_limit is not None:
                    x = np.fmax(x, lower_limit)
                if upper_limit is not None:
                    x = np.fmin(x, upper_limit)
                if x_mask is not None:
                    x *= x_mask

            if self.tolerance is not None:
                res = self.compute_residual(A, b, x, A_num_rows=A_num_rows, b_mask=b_mask)
                info.residuals[ii] = np.linalg.norm(res)

                if self.tolerance > info.residuals[ii]:
                    break

        return x, info


class MLEM(Solver):
    """
    Initialize the MLEM solver class.

    This class implements the Maximul Likelihood Expectation Maximization (MLEM) algorithm.

    Parameters
    ----------
    verbose : bool, optional
        Turn on verbose output. The default is False.
    leave_progress: bool, optional
        Leave the progress bar after the computation is finished. The default is True.
    tolerance : Optional[float], optional
        Tolerance on the data residual for computing when to stop iterations.
        The default is None.
    regularizer : Sequence[regularizers.BaseRegularizer] | regularizers.BaseRegularizer | None, optional
        Regularizer to be used. The default is None.
    data_term : Union[str, data_terms.DataFidelityBase], optional
        Data fidelity term for computing the data residual. The default is "l2".
    data_term_test : Optional[data_terms.DataFidelityBase], optional
        The data fidelity to be used for the test set.
        If None, it will use the same as for the rest of the data.
        The default is None.
    """

    def __init__(
        self,
        verbose: bool = False,
        leave_progress: bool = True,
        tolerance: Optional[float] = None,
        regularizer: Union[Sequence[regularizers.BaseRegularizer], regularizers.BaseRegularizer, None] = None,
        data_term: Union[str, data_terms.DataFidelityBase] = "kl",
        data_term_test: Union[str, data_terms.DataFidelityBase, None] = None,
    ):
        super().__init__(
            verbose=verbose,
            leave_progress=leave_progress,
            tolerance=tolerance,
            data_term=data_term,
            data_term_test=data_term_test,
        )
        self.regularizer = self._initialize_regularizer(regularizer)

    def info(self) -> str:
        """
        Return the MLEM info.

        Returns
        -------
        str
            info string.
        """
        return Solver.info(self) + f"(B:{self.data_term.background:g})" if self.data_term.background is not None else ""

    def __call__(  # noqa: C901
        self,
        A: operators.BaseTransform,
        b: NDArrayFloat,
        iterations: int,
        x0: Optional[NDArrayFloat] = None,
        lower_limit: Union[float, NDArrayFloat, None] = None,
        upper_limit: Union[float, NDArrayFloat, None] = None,
        x_mask: Optional[NDArrayFloat] = None,
        b_mask: Optional[NDArrayFloat] = None,
        b_test_mask: Optional[NDArrayFloat] = None,
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """
        Reconstruct the data, using the MLEM algorithm.

        Parameters
        ----------
        A : BaseTransform
            Projection operator.
        b : NDArrayFloat
            Data to reconstruct.
        iterations : int
            Number of iterations.
        x0 : Optional[NDArrayFloat], optional
            Initial solution. The default is None.
        lower_limit : Union[float, NDArrayFloat], optional
            Lower clipping value. The default is None.
        upper_limit : Union[float, NDArrayFloat], optional
            Upper clipping value. The default is None.
        x_mask : Optional[NDArrayFloat], optional
            Solution mask. The default is None.
        b_mask : Optional[NDArrayFloat], optional
            Data mask. The default is None.
        b_test_mask : Optional[NDArrayFloat], optional
            Test data mask. The default is None.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction, and the residuals.
        """
        b = np.array(b)

        (b_mask, b_test_mask) = self._initialize_b_masks(b, b_mask, b_test_mask)

        # Back-projection diagonal re-scaling
        b_ones = np.ones_like(b)
        if b_mask is not None:
            b_ones *= b_mask
        tau = A.T(b_ones)

        # Forward-projection diagonal re-scaling
        x_ones = np.ones_like(tau)
        if x_mask is not None:
            x_ones *= x_mask
        sigma = np.abs(A(x_ones))
        sigma[(sigma / np.max(sigma)) < 1e-5] = 1
        sigma = 1 / sigma

        if x0 is None:
            x = np.ones_like(tau)
        else:
            x = np.array(x0).copy()
        if x_mask is not None:
            x *= x_mask

        self.data_term.assign_data(b)

        info = SolutionInfo(self.info(), max_iterations=iterations, tolerance=self.tolerance)

        if b_test_mask is not None or self.tolerance is not None:
            Ax = A(x)

            if b_test_mask is not None:
                if self.data_term_test.background != self.data_term.background:
                    print("WARNING - the data_term and and data_term_test should have the same background. Making them equal.")
                    self.data_term_test.background = self.data_term.background
                self.data_term_test.assign_data(b)

                res_test_0 = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                info.residual0_cv = self.data_term_test.compute_residual_norm(res_test_0)

            if self.tolerance is not None:
                res_0 = self.data_term.compute_residual(Ax, mask=b_mask)
                info.residual0 = self.data_term.compute_residual_norm(res_0)

        reg_info = "".join(["-" + r.info().upper() for r in self.regularizer])
        algo_info = f"- Performing {self.upper()}-{self.data_term.upper()}{reg_info} iterations: "

        for ii in tqdm(range(iterations), desc=algo_info, disable=(not self.verbose), leave=self.leave_progress):
            info.iterations += 1

            # The MLEM update
            Ax = A(x)

            if b_test_mask is not None:
                res_test = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                info.residuals_cv[ii] = self.data_term_test.compute_residual_norm(res_test)

            if self.tolerance is not None:
                res = self.data_term.compute_residual(Ax, mask=b_mask)
                info.residuals[ii] = self.data_term.compute_residual_norm(res)
                if self.tolerance > info.residuals[ii]:
                    break

            if self.data_term.background is not None:
                Ax = Ax + self.data_term.background
            Ax = Ax.clip(eps, None)

            upd = A.T(b / Ax)
            x *= upd / tau

            if lower_limit is not None or upper_limit is not None:
                x = x.clip(lower_limit, upper_limit)
            if x_mask is not None:
                x *= x_mask

        return x, info


class SIRT(Solver):
    """
    Initialize the SIRT solver class.

    This class implements the Simultaneous Iterative Reconstruction Technique (SIRT) algorithm.

    Parameters
    ----------
    verbose : bool, optional
        Turn on verbose output. The default is False.
    leave_progress: bool, optional
        Leave the progress bar after the computation is finished. The default is True.
    tolerance : Optional[float], optional
        Tolerance on the data residual for computing when to stop iterations.
        The default is None.
    relaxation : float, optional
        The relaxation length. The default is 1.95.
    regularizer : Sequence[regularizers.BaseRegularizer] | regularizers.BaseRegularizer | None, optional
        Regularizer to be used. The default is None.
    data_term : Union[str, data_terms.DataFidelityBase], optional
        Data fidelity term for computing the data residual. The default is "l2".
    data_term_test : Optional[data_terms.DataFidelityBase], optional
        The data fidelity to be used for the test set.
        If None, it will use the same as for the rest of the data.
        The default is None.
    """

    def __init__(
        self,
        verbose: bool = False,
        leave_progress: bool = True,
        relaxation: float = 1.95,
        tolerance: Optional[float] = None,
        regularizer: Union[Sequence[regularizers.BaseRegularizer], regularizers.BaseRegularizer, None] = None,
        data_term: Union[str, data_terms.DataFidelityBase] = "l2",
        data_term_test: Union[str, data_terms.DataFidelityBase, None] = None,
    ):
        super().__init__(
            verbose=verbose,
            leave_progress=leave_progress,
            relaxation=relaxation,
            tolerance=tolerance,
            data_term=data_term,
            data_term_test=data_term_test,
        )
        self.regularizer = self._initialize_regularizer(regularizer)

    def info(self) -> str:
        """
        Return the SIRT info.

        Returns
        -------
        str
            SIRT info string.
        """
        reg_info = "".join(["-" + r.info().upper() for r in self.regularizer])
        return Solver.info(self) + "-" + self.data_term.info() + reg_info

    def __call__(  # noqa: C901
        self,
        A: operators.BaseTransform,
        b: NDArrayFloat,
        iterations: int,
        x0: Optional[NDArrayFloat] = None,
        lower_limit: Union[float, NDArrayFloat, None] = None,
        upper_limit: Union[float, NDArrayFloat, None] = None,
        x_mask: Optional[NDArrayFloat] = None,
        b_mask: Optional[NDArrayFloat] = None,
        b_test_mask: Optional[NDArrayFloat] = None,
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """
        Reconstruct the data, using the SIRT algorithm.

        Parameters
        ----------
        A : BaseTransform
            Projection operator.
        b : NDArrayFloat
            Data to reconstruct.
        iterations : int
            Number of iterations.
        x0 : Optional[NDArrayFloat], optional
            Initial solution. The default is None.
        lower_limit : Union[float, NDArrayFloat], optional
            Lower clipping value. The default is None.
        upper_limit : Union[float, NDArrayFloat], optional
            Upper clipping value. The default is None.
        x_mask : Optional[NDArrayFloat], optional
            Solution mask. The default is None.
        b_mask : Optional[NDArrayFloat], optional
            Data mask. The default is None.
        b_test_mask : Optional[NDArrayFloat], optional
            Test data mask. The default is None.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction, and the residuals.
        """
        b = np.array(b)

        (b_mask, b_test_mask) = self._initialize_b_masks(b, b_mask, b_test_mask)

        # Back-projection diagonal re-scaling
        b_ones = np.ones_like(b)
        if b_mask is not None:
            b_ones *= b_mask
        tau = np.abs(A.T(b_ones))
        for reg in self.regularizer:
            tau += reg.initialize_sigma_tau(tau)
        tau[(tau / np.max(tau)) < 1e-5] = 1
        tau = self.relaxation / tau

        # Forward-projection diagonal re-scaling
        x_ones = np.ones_like(tau)
        if x_mask is not None:
            x_ones *= x_mask
        sigma = np.abs(A(x_ones))
        sigma[(sigma / np.max(sigma)) < 1e-5] = 1
        sigma = 1 / sigma

        if x0 is None:
            x = np.zeros_like(x_ones)
        else:
            x = np.array(x0).copy()

        self.data_term.assign_data(b, sigma)

        info = SolutionInfo(self.info(), max_iterations=iterations, tolerance=self.tolerance)

        if b_test_mask is not None or self.tolerance is not None:
            Ax = A(x)

            res_0 = self.data_term.compute_residual(Ax, mask=b_mask)
            info.residual0 = self.data_term.compute_residual_norm(res_0)

            if b_test_mask is not None:
                if self.data_term_test.background != self.data_term.background:
                    print("WARNING - the data_term and and data_term_test should have the same background. Making them equal.")
                    self.data_term_test.background = self.data_term.background
                self.data_term_test.assign_data(b, sigma)

                res_test_0 = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                info.residual0_cv = self.data_term_test.compute_residual_norm(res_test_0)

        reg_info = "".join(["-" + r.info().upper() for r in self.regularizer])
        algo_info = f"- Performing {self.upper()}-{self.data_term.upper()}{reg_info} iterations: "

        for ii in tqdm(range(iterations), desc=algo_info, disable=(not self.verbose), leave=self.leave_progress):
            info.iterations += 1

            Ax = A(x)
            res = self.data_term.compute_residual(Ax, mask=b_mask)

            if b_test_mask is not None or self.tolerance is not None:
                info.residuals[ii] = self.data_term.compute_residual_norm(res)

                if b_test_mask is not None:
                    res_test = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                    info.residuals_cv[ii] = self.data_term_test.compute_residual_norm(res_test)

                if self.tolerance is not None and self.tolerance > info.residuals[ii]:
                    if self.verbose:
                        print(f"Residual reached the desired tolerance of {self.tolerance}. Ending iterations..")
                    break

            q = [reg.initialize_dual() for reg in self.regularizer]
            for q_r, reg in zip(q, self.regularizer):
                reg.update_dual(q_r, x)
                reg.apply_proximal(q_r)

            upd = A.T(res * sigma)
            for q_r, reg in zip(q, self.regularizer):
                upd -= reg.compute_update_primal(q_r)
            x += upd * tau

            if lower_limit is not None or upper_limit is not None:
                x = x.clip(lower_limit, upper_limit)
            if x_mask is not None:
                x *= x_mask

        return x, info


class PDHG(Solver):
    """
    Initialize the PDHG solver class.

    PDHG stands for primal-dual hybrid gradient algorithm from Chambolle and Pock.

    Parameters
    ----------
    verbose : bool, optional
        Turn on verbose output. The default is False.
    leave_progress: bool, optional
        Leave the progress bar after the computation is finished. The default is True.
    tolerance : Optional[float], optional
        Tolerance on the data residual for computing when to stop iterations.
        The default is None.
    relaxation : float, optional
        The relaxation length. The default is 0.95.
    regularizer : Sequence[regularizers.BaseRegularizer] | regularizers.BaseRegularizer | None, optional
        Regularizer to be used. The default is None.
    data_term : Union[str, data_terms.DataFidelityBase], optional
        Data fidelity term for computing the data residual. The default is "l2".
    data_term_test : Optional[data_terms.DataFidelityBase], optional
        The data fidelity to be used for the test set.
        If None, it will use the same as for the rest of the data.
        The default is None.
    """

    def __init__(
        self,
        verbose: bool = False,
        leave_progress: bool = True,
        tolerance: Optional[float] = None,
        relaxation: float = 0.95,
        regularizer: Union[Sequence[regularizers.BaseRegularizer], regularizers.BaseRegularizer, None] = None,
        data_term: Union[str, data_terms.DataFidelityBase] = "l2",
        data_term_test: Union[str, data_terms.DataFidelityBase, None] = None,
    ):
        super().__init__(
            verbose=verbose,
            leave_progress=leave_progress,
            relaxation=relaxation,
            tolerance=tolerance,
            data_term=data_term,
            data_term_test=data_term_test,
        )
        self.regularizer = self._initialize_regularizer(regularizer)

    def info(self) -> str:
        """
        Return the PDHG info.

        Returns
        -------
        str
            PDHG info string.
        """
        reg_info = "".join(["-" + r.info().upper() for r in self.regularizer])
        return Solver.info(self) + "-" + self.data_term.info() + reg_info

    @staticmethod
    def _initialize_data_fidelity_function(data_term: Union[str, data_terms.DataFidelityBase]):
        if isinstance(data_term, str):
            if data_term.lower() == "l2":
                return data_terms.DataFidelity_l2()
            if data_term.lower() == "l1":
                return data_terms.DataFidelity_l1()
            if data_term.lower() == "kl":
                return data_terms.DataFidelity_KL()
            else:
                raise ValueError(f'Unknown data term: "{data_term}", accepted terms are: "l2" | "l1" | "kl".')
        else:
            return cp.deepcopy(data_term)

    def power_method(
        self, A: operators.BaseTransform, b: NDArrayFloat, iterations: int = 5
    ) -> tuple[np.floating, Sequence[int], DTypeLike]:
        """
        Compute the l2-norm of the operator A, with the power method.

        Parameters
        ----------
        A : BaseTransform
            Operator whose l2-norm needs to be computed.
        b : NDArrayFloat
            The data vector.
        iterations : int, optional
            Number of power method iterations. The default is 5.

        Returns
        -------
        Tuple[float, Tuple[int], DTypeLike]
            The l2-norm of A, and the shape and type of the solution.
        """
        x: NDArrayFloat = np.array(np.random.rand(*b.shape))
        x = x.astype(b.dtype)
        x /= np.linalg.norm(x)
        x = A.T(x)

        x_norm = np.linalg.norm(x)
        L = x_norm

        for _ in range(iterations):
            x /= x_norm
            x = A.T(A(x))

            x_norm = np.linalg.norm(x)
            L = np.sqrt(x_norm)

        return (L, x.shape, x.dtype)

    def _get_data_sigma_tau_unpreconditioned(self, A: operators.BaseTransform, b: NDArrayFloat):
        (L, x_shape, x_dtype) = self.power_method(A, b)
        tau = L

        dummy_x = np.empty(x_shape, dtype=x_dtype)
        for reg in self.regularizer:
            tau += reg.initialize_sigma_tau(dummy_x)

        tau = self.relaxation / tau
        sigma = self.relaxation / L
        return (x_shape, x_dtype, sigma, tau)

    def __call__(  # noqa: C901
        self,
        A: operators.BaseTransform,
        b: NDArrayFloat,
        iterations: int,
        x0: Optional[NDArrayFloat] = None,
        lower_limit: Union[float, NDArrayFloat, None] = None,
        upper_limit: Union[float, NDArrayFloat, None] = None,
        x_mask: Optional[NDArrayFloat] = None,
        b_mask: Optional[NDArrayFloat] = None,
        b_test_mask: Optional[NDArrayFloat] = None,
        precondition: bool = True,
    ) -> tuple[NDArrayFloat, SolutionInfo]:
        """
        Reconstruct the data, using the PDHG algorithm.

        Parameters
        ----------
        A : BaseTransform
            Projection operator.
        b : NDArrayFloat
            Data to reconstruct.
        iterations : int
            Number of iterations.
        x0 : Optional[NDArrayFloat], optional
            Initial solution. The default is None.
        lower_limit : Union[float, NDArrayFloat], optional
            Lower clipping value. The default is None.
        upper_limit : Union[float, NDArrayFloat], optional
            Upper clipping value. The default is None.
        x_mask : Optional[NDArrayFloat], optional
            Solution mask. The default is None.
        b_mask : Optional[NDArrayFloat], optional
            Data mask. The default is None.
        b_test_mask : Optional[NDArrayFloat], optional
            Test data mask. The default is None.
        precondition : bool, optional
            Whether to use the preconditioned version of the algorithm. The default is True.

        Returns
        -------
        Tuple[NDArrayFloat, SolutionInfo]
            The reconstruction, and the residuals.
        """
        b = np.array(b)

        if precondition:
            try:
                At_abs = A.T.absolute()
                A_abs = A.absolute()
            except AttributeError:
                print(A)
                print("WARNING: Turning off preconditioning because system matrix does not support absolute")
                precondition = False

        (b_mask, b_test_mask) = self._initialize_b_masks(b, b_mask, b_test_mask)

        if precondition:
            tau = np.ones_like(b)
            if b_mask is not None:
                tau *= b_mask
            tau = np.abs(At_abs(tau))
            for reg in self.regularizer:
                tau += reg.initialize_sigma_tau(tau)
            tau[(tau / np.max(tau)) < 1e-5] = 1
            tau = self.relaxation / tau

            x_shape = tau.shape
            x_dtype = tau.dtype

            sigma = np.ones_like(tau)
            if x_mask is not None:
                sigma *= x_mask
            sigma = np.abs(A_abs(sigma))
            sigma[(sigma / np.max(sigma)) < 1e-5] = 1
            sigma = self.relaxation / sigma
        else:
            (x_shape, x_dtype, sigma, tau) = self._get_data_sigma_tau_unpreconditioned(A, b)

        if x0 is None:
            x0 = np.zeros(x_shape, dtype=x_dtype)
        else:
            x0 = np.array(x0).copy()
        x = x0
        x_relax = x.copy()

        self.data_term.assign_data(b, sigma)
        p = self.data_term.initialize_dual()

        q = [reg.initialize_dual() for reg in self.regularizer]

        info = SolutionInfo(self.info(), max_iterations=iterations, tolerance=self.tolerance)

        if b_test_mask is not None or self.tolerance is not None:
            Ax = A(x)

            res_0 = self.data_term.compute_residual(Ax, mask=b_mask)
            info.residual0 = self.data_term.compute_residual_norm(res_0)

            if b_test_mask is not None:
                if self.data_term_test.background != self.data_term.background:
                    print("WARNING - the data_term and and data_term_test should have the same background. Making them equal.")
                    self.data_term_test.background = self.data_term.background
                self.data_term_test.assign_data(b, sigma)

                res_test_0 = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                info.residual0_cv = self.data_term_test.compute_residual_norm(res_test_0)

        reg_info = "".join(["-" + r.info().upper() for r in self.regularizer])
        algo_info = f"- Performing {self.upper()}-{self.data_term.upper()}{reg_info} iterations: "

        for ii in tqdm(range(iterations), desc=algo_info, disable=(not self.verbose), leave=self.leave_progress):
            info.iterations += 1

            Ax_rlx = A(x_relax)
            self.data_term.update_dual(p, Ax_rlx)
            self.data_term.apply_proximal(p)

            if b_mask is not None:
                p *= b_mask

            for q_r, reg in zip(q, self.regularizer):
                reg.update_dual(q_r, x_relax)
                reg.apply_proximal(q_r)

            upd = A.T(p)
            for q_r, reg in zip(q, self.regularizer):
                upd += reg.compute_update_primal(q_r)
            x_new = x - upd * tau

            if lower_limit is not None or upper_limit is not None:
                x_new = x_new.clip(lower_limit, upper_limit)
            if x_mask is not None:
                x_new *= x_mask

            x_relax = x_new + (x_new - x)
            x = x_new

            if b_test_mask is not None or self.tolerance is not None:
                Ax = A(x)
                res = self.data_term.compute_residual(Ax, mask=b_mask)
                info.residuals[ii] = self.data_term.compute_residual_norm(res)

                if b_test_mask is not None:
                    res_test = self.data_term_test.compute_residual(Ax, mask=b_test_mask)
                    info.residuals_cv[ii] = self.data_term_test.compute_residual_norm(res_test)

                if self.tolerance is not None and self.tolerance > info.residuals[ii]:
                    if self.verbose:
                        print(f"Residual reached the desired tolerance of {self.tolerance}. Ending iterations..")
                    break

        return x, info



"""
Pre-processing routines.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from collections.abc import Sequence
from typing import Optional, Union
import matplotlib.pyplot as plt
import numpy as np
import pywt
import scipy.ndimage as spimg
import skimage.transform as skt
from numpy.polynomial import Polynomial
from numpy.typing import DTypeLike, NDArray
from skimage.measure import block_reduce


eps = np.finfo(np.float32).eps


def pad_sinogram(
    sinogram: NDArray, width: Union[int, Sequence[int], NDArray], pad_axis: int = -1, mode: str = "edge", **kwds
) -> NDArray:
    """
    Pad the sinogram.

    Parameters
    ----------
    sinogram : NDArray
        The sinogram to pad.
    width : Union[int, Sequence[int]]
        The width of the padding. Normally, it should either be an int or a tuple(int, int).
    pad_axis : int, optional
        The axis to pad. The default is -1.
    mode : str, optional
        The padding type (from numpy.pad). The default is "edge".
    **kwds :
        The numpy.pad arguments.

    Returns
    -------
    NDArray
        The padded sinogram.
    """
    width = np.array(width, ndmin=1)

    pad_size = np.zeros((len(sinogram.shape), len(width)), dtype=int)
    pad_size[pad_axis, :] = width

    return np.pad(sinogram, pad_size, mode=mode.lower(), **kwds)  # type: ignore


def apply_flat_field(
    projs_wvu: NDArray,
    flats_wvu: NDArray,
    darks_wvu: Optional[NDArray] = None,
    crop: Union[NDArray, Sequence[int], None] = None,
    cap_intensity: Optional[float] = None,
    dtype: DTypeLike = np.float32,
) -> NDArray:
    """
    Apply flat field.

    Parameters
    ----------
    projs : NDArray
        Projections.
    flats : NDArray
        Flat fields.
    darks : Optional[NDArray], optional
        Dark noise. The default is None.
    crop : Optional[Sequence[int]], optional
        Crop region. The default is None.
    cap_intensity: float | None, optional
        Cap the intensity to a given value. The default is None.
    dtype : DTypeLike, optional
        Data type of the processed data. The default is np.float32.

    Returns
    -------
    NDArray
        Flat-field corrected and linearized projections.
    """
    projs_wvu = np.ascontiguousarray(projs_wvu, dtype=dtype)
    flats_wvu = np.ascontiguousarray(flats_wvu, dtype=dtype)

    if crop is not None:
        projs_wvu = projs_wvu[..., crop[0] : crop[2], crop[1] : crop[3]]
        flats_wvu = flats_wvu[..., crop[0] : crop[2], crop[1] : crop[3]]
        if darks_wvu is not None:
            darks_wvu = darks_wvu[..., crop[0] : crop[2], crop[1] : crop[3]]

    if darks_wvu is not None:
        darks_wvu = np.ascontiguousarray(darks_wvu, dtype=dtype)
        projs_wvu = projs_wvu - darks_wvu
        flats_wvu = flats_wvu - darks_wvu

    if flats_wvu.ndim == 3:
        flats_wvu = np.mean(flats_wvu, axis=0)

    projs_wvu = projs_wvu / flats_wvu

    if cap_intensity is not None:
        projs_wvu = np.fmin(projs_wvu, cap_intensity)

    return projs_wvu


def apply_minus_log(projs: NDArray, lower_limit: float = -np.inf) -> NDArray:
    """
    Apply -log.

    Parameters
    ----------
    projs : NDArray
        Projections.

    Returns
    -------
    NDArray
        Linearized projections.
    """
    return np.fmax(-np.log(projs), lower_limit)


def rotate_proj_stack(data_vwu: NDArray, rot_angle_deg: float) -> NDArray:
    """
    Rotate the projection stack.

    Parameters
    ----------
    data_vwu : NDArray
        The projection stack, with dimensions [v, w, u] (vertical, omega / sample rotation, horizontal).
    rot_angle_deg : float
        The rotation angle in degrees.

    Returns
    -------
    NDArray
        The rotated projection stack.
    """
    data_vwu_r = np.empty_like(data_vwu)
    for ii in range(data_vwu.shape[-2]):
        data_vwu_r[:, ii, :] = skt.rotate(data_vwu[:, ii, :], -rot_angle_deg, clip=False)
    return data_vwu_r


def shift_proj_stack(data_vwu: NDArray, shifts: NDArray, use_fft: bool = False) -> NDArray:
    """Shift each projection in a stack of projections, by projection dependent shifts.

    Parameters
    ----------
    data_vwu : NDArray
        The projection stack
    shifts : NDArray
        The shifts
    use_fft : bool, optional
        Whether to use fft shift or not, by default False

    Returns
    -------
    NDArray
        The shifted stack
    """
    new_data = np.empty_like(data_vwu)
    for ii in range(data_vwu.shape[-2]):
        if use_fft:
            img = data_vwu[..., ii, :]
            img_f = np.fft.rfftn(img)
            img_f = spimg.fourier_shift(img_f, shifts[..., ii], n=img.shape[-1])
            new_data[..., ii, :] = np.fft.irfftn(img_f)
        else:
            new_data[..., ii, :] = spimg.shift(data_vwu[..., ii, :], shifts[..., ii], order=1, mode="nearest")

    return new_data


def bin_imgs(
    imgs: NDArray, binning: Union[int, float], axes: Sequence[int] = (-2, -1), auto_crop: bool = False, verbose: bool = True
) -> NDArray:
    """Bin a stack of images.

    Parameters
    ----------
    imgs : NDArray
        The stack of images.
    binning : int | float
        The binning factor.
    auto_crop : bool, optional
        Whether to automatically crop the images to match, by default False
    verbose : bool, optional
        Whether to print the image shapes, by default True

    Returns
    -------
    NDArray
        The binned images
    """
    if auto_crop:
        imgs_shape = np.array(imgs.shape)
        excess_pixels = (imgs_shape[list(axes)] % binning).astype(int)
        crop_vu = (excess_pixels - excess_pixels // 2, imgs_shape[list(axes)] - excess_pixels // 2)
        slicing = [slice(None)] * len(imgs_shape)
        for ii, ax in enumerate(axes):
            if excess_pixels[ii] > 0:
                slicing[ax] = slice(crop_vu[0][ii], crop_vu[1][ii])
        imgs = imgs[tuple(slicing)]

        if verbose:
            print(f"Auto-cropping {crop_vu}: {imgs_shape} => {imgs.shape}")

    imgs_shape = imgs.shape

    if isinstance(binning, int):
        binning_shape = np.ones_like(imgs_shape)
        for ax in axes:
            binning_shape[ax] = binning
        imgs = block_reduce(imgs, tuple(binning_shape), np.mean)
        binned_shape = imgs.shape
    else:
        imgs = imgs.reshape([-1, *imgs_shape[-2:]])
        imgs = skt.rescale(imgs, 1 / binning, channel_axis=0 if len(imgs_shape) > 2 else None)
        binned_shape = [*imgs_shape[:-2], *imgs.shape[-2:]]
        imgs = imgs.reshape(binned_shape)

    if verbose:
        print(f"Binning {binning}: {imgs_shape} => {binned_shape}")

    return imgs


def background_from_margin(
    data_vwu: NDArray, margin: Union[int, Sequence[int], NDArray[np.integer]] = 4, poly_order: int = 0, plot: bool = False
) -> NDArray:
    """Compute background of the projection data, from the margins of the projections.

    Parameters
    ----------
    data_vwu : NDArray
        The projection data in the format [V]WU.
    margin : int | Sequence[int] | NDArray[np.integer], optional
        The size of the margin, by default 4
    poly_order : int, optional
        The order of the interpolation polynomial, by default 0

    Returns
    -------
    NDArray
        The computed background.

    Raises
    ------
    NotImplementedError
        Different margins per line are not supported, at the moment.
    ValueError
        In case the margins ar larger than the image size in U.
    """
    data_shape_u = data_vwu.shape[-1]
    data_shape_w = data_vwu.shape[-2]
    margin = np.array(margin, dtype=int, ndmin=1)
    if margin.size == 1:
        margin = np.tile(margin, [*np.ones(margin.ndim - 1), 2])
    if margin.ndim > 1:
        raise NotImplementedError("Complex masks support has not been implemented, yet.")
    if margin.sum() > data_shape_u:
        raise ValueError(f"Margin size ({margin}) should be smaller than the image size in U ({data_shape_u})")
    if poly_order > 0 and np.any(margin == 0):
        print("WARNING: parameter `poly_order` cannot be greater than 0 if one of the margins is 0")
        poly_order = 0

    if poly_order > 0:
        ydata = np.concatenate([data_vwu[..., : margin[0]], data_vwu[..., -margin[1] :]], axis=-1)
        xdata = np.concatenate([np.arange(0, margin[0]), np.arange(data_shape_u - margin[1], data_shape_u)])

        if data_vwu.ndim > 2:
            ydata = ydata.mean(axis=-3)

        background = np.empty([data_shape_w, data_shape_u], dtype=data_vwu.dtype)
        for ii_w in range(data_shape_w):
            poly = Polynomial.fit(xdata, ydata[ii_w], deg=poly_order)
            background[ii_w, :] = poly(np.arange(data_shape_u))

        if plot:
            fig, axs = plt.subplots(1, 1)
            axs.plot(background[0])
            axs.scatter(xdata, ydata[0])
            axs.grid()
            axs.set_ylim(0)
            fig.tight_layout()
            plt.show(block=False)

        if data_vwu.ndim > 2:
            background = np.tile(background[None, ...], [data_vwu.shape[-3], 1, 1])
        return background
    else:
        sum_vals = data_vwu[..., : margin[0]].sum(axis=-1) + data_vwu[..., -margin[1] :].sum(axis=-1)
        background: NDArray = sum_vals / margin.sum(axis=-1)

        if data_vwu.ndim > 2:
            background = background.mean(axis=-2, keepdims=True)

        return np.tile(background[..., None], [*np.ones(background.ndim, dtype=int), data_shape_u])


def destripe_wlf_vwu(
    data: NDArray,
    sigma: float = 0.005,
    level: int = 1,
    wavelet: str = "bior2.2",
    angle_axis: int = -2,
    other_axes: Union[Sequence[int], NDArray, None] = None,
) -> NDArray:
    """Remove stripes from sinogram, using the Wavelet-Fourier method.

    Parameters
    ----------
    data : NDArray
        The data to de-stripe
    sigma : float, optional
        Fourier space filter coefficient, by default 0.005
    level : int, optional
        The wavelet level to use, by default 1
    wavelet : str, optional
        The type of wavelet to use, by default "bior2.2"
    angle_axis : int, optional
        The axis of the Fourier transform, by default -2
    other_axes : Union[Sequence[int], NDArray, None], optional
        The axes of the wavelet decomposition, by default None

    Returns
    -------
    NDArray
        The de-striped data.
    """
    if other_axes is None:
        other_axes = np.arange(-data.ndim, 0)
    else:
        other_axes = np.array(other_axes)

    if angle_axis is other_axes:
        other_axes = np.delete(other_axes, angle_axis)

    level_power = 2**level

    data_shape = np.array(data.shape)
    target_shape = data_shape.copy()
    target_shape[list(other_axes)] = np.ceil(data_shape[list(other_axes)] / level_power) * level_power
    diff_size = target_shape - data_shape
    padding = np.stack((diff_size - diff_size // 2, diff_size // 2), axis=-1)
    data = np.pad(data, pad_width=padding, mode="edge")

    coeffs = pywt.swtn(data, wavelet=wavelet, axes=other_axes, level=level)
    for ii_l in range(level):
        for wl_label, coeffs_l_wl in coeffs[ii_l].items():
            if wl_label == "a" * len(other_axes):
                continue
            coeff_f = np.fft.rfft(coeffs_l_wl, axis=angle_axis)
            filt_f = 1 - np.exp(-(np.fft.rfftfreq(coeffs_l_wl.shape[angle_axis]) ** 2) / (2 * sigma**2))
            coeff_f *= filt_f[:, None]
            coeffs[ii_l][wl_label] = np.fft.irfft(coeff_f, axis=angle_axis, n=coeffs_l_wl.shape[angle_axis])

    data = pywt.iswtn(coeffs, wavelet=wavelet, axes=other_axes)
    slicing = [slice(padding[ii, 0], data.shape[ii] - padding[ii, 1]) for ii in range(data.ndim)]

    return data[tuple(slicing)]


def compute_eigen_flats(
    trans_wvu: NDArray,
    flats_wvu: Optional[NDArray] = None,
    darks_wvu: Optional[NDArray] = None,
    ndim: int = 2,
    plot: bool = False,
) -> tuple[NDArray, NDArray]:
    """Compute the eigen flats of a stack of transmission images.

    Parameters
    ----------
    trans : NDArray
        The stack of transmission images.
    flats : NDArray
        The flats without sample.
    darks : NDArray
        The darks.
    ndim : int, optional
        The number of dimensions of the images, by default 2
    plot : bool, optional
        Whether to plot the results, by default False

    Returns
    -------
    Tuple[NDArray, NDArray]
        The decomposition of the transmissions of the sample and the flats.
    """
    trans_shape = trans_wvu.shape
    trans_num = np.prod(trans_shape[:-ndim])
    img_shape = trans_shape[-ndim:]

    stack_imgs = [trans_wvu.reshape([-1, *img_shape])]

    if flats_wvu is not None:
        stack_imgs.append(flats_wvu.reshape([-1, *img_shape]))

    stack_imgs = np.concatenate(stack_imgs)

    if darks_wvu is not None:
        if darks_wvu.ndim > 2:
            darks_wvu = darks_wvu.mean(axis=tuple(np.arange(darks_wvu.ndim - 2)))
        stack_imgs = np.fmax(stack_imgs - darks_wvu, np.finfo(np.float32).eps)

    stack_imgs = stack_imgs.reshape([-1, np.prod(img_shape)]).transpose()
    stack_imgs = np.log(stack_imgs)

    mat_u, sigma, mat_v_h = np.linalg.svd(stack_imgs, full_matrices=False)

    eigen_projs: NDArray = (mat_u[..., 1:] * sigma[..., None, 1:]) @ mat_v_h[..., 1:, :]
    eigen_projs = np.exp(eigen_projs)
    eigen_projs = eigen_projs.transpose().reshape([-1, *img_shape])[:trans_num]

    eigen_flats: NDArray = (mat_u[..., 0:1:] * sigma[..., None, 0:1:]) @ mat_v_h[..., 0:1:, :]
    eigen_flats = np.exp(eigen_flats)
    eigen_flats = eigen_flats.transpose().reshape([-1, *img_shape])[:trans_num]

    if plot:
        fig, axs = plt.subplots(1, 3, figsize=[10, 3.75])
        axs[0].plot(sigma)
        axs[0].grid()
        axs[0].set_title("Singular values")
        axs[1].imshow(mat_u[:, 0].reshape(img_shape))
        axs[1].set_title("Highest value component")
        axs[2].plot(eigen_flats.mean(axis=(-2, -1)).flatten())
        axs[2].grid()
        axs[2].set_title("Eigen intensities")
        fig.tight_layout()
        plt.show(block=False)

    return eigen_projs.reshape(trans_shape), eigen_flats.reshape(trans_shape)



#!/usr/bin/env python3
"""
Fitting routines.

Created on Tue May 17 12:11:58 2022

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from typing import Optional, Union, Sequence

import numpy as np
from numpy.polynomial import Polynomial
import scipy.ndimage as spimg
import scipy.optimize as spopt
from numpy.typing import ArrayLike, NDArray
from skimage.transform import warp_polar
from skimage.filters import window
from scipy.optimize import minimize
import matplotlib.pyplot as plt

NDArrayFloat = NDArray[np.floating]


eps = np.finfo(np.float32).eps


def fit_shifts_u_sad(
    data_wu: NDArrayFloat,
    proj_wu: NDArrayFloat,
    search_range: int = 16,
    pad_u: bool = False,
    error_norm: int = 1,
    decimals: int = 2,
) -> NDArrayFloat:
    """
    Find the U shifts between two sets of lines, by means of the sum-of-absolute-difference (SAD).

    Parameters
    ----------
    data_wu : NDArrayFloat
        The reference data.
    proj_wu : NDArrayFloat
        The other data.
    search_rage : int, optional
        The range in pixels of the search, by default 16
    error_norm : int, optional
        The error norm to use, by default 1
    decimals : int, optional
        The precision of the result, by default 2

    Returns
    -------
    NDArrayFloat
        A list of one shift for each row.
    """
    if pad_u:
        padding = np.zeros((len(data_wu.shape), 2), dtype=int)
        padding[-1, :] = (search_range, search_range)
        pad_data_wu = np.pad(data_wu, pad_width=padding, mode="edge")
        pad_proj_wu = np.pad(proj_wu, pad_width=padding, mode="constant")

    fft_proj_wu = np.fft.fft2(pad_proj_wu)

    num_shifts = search_range * 2 + 1
    shift_coords = np.fft.fftfreq(num_shifts, 1 / num_shifts)

    diffs = np.empty((data_wu.shape[-2], len(shift_coords)))
    for ii, s in enumerate(shift_coords):
        shifted_proj_wu = np.fft.ifft2(spimg.fourier_shift(fft_proj_wu, (0, s))).real
        diffs[:, ii] = np.linalg.norm(pad_data_wu - shifted_proj_wu, axis=-1, ord=error_norm)

    f_vals, f_h = extract_peak_regions_1d(-diffs, axis=-1, cc_coords=shift_coords)
    shifts_vu = f_h[1, :]

    if decimals > 0:
        shifts_vu += refine_max_position_1d(f_vals, decimals=decimals)

    return shifts_vu


def fit_shifts_vu_xc(
    data_vwu: NDArrayFloat,
    proj_vwu: NDArrayFloat,
    pad_u: bool = False,
    normalize_fourier: bool = False,
    margin: int = 0,
    use_rfft: bool = True,
    stack_axis: int = -2,
    decimals: int = 2,
) -> NDArrayFloat:
    """
    Find the VU shifts of the projected data, through cross-correlation.

    Parameters
    ----------
    data_vwu : NDArrayFloat
        The collected projection data.
    proj_vwu : NDArrayFloat
        The forward-projected images from the reconstruction.
    pad_u : bool, optional
        Pad the u coordinate. The default is False.
    normalize_fourier : bool, optional
        Whether to normalize the Fourier representation of the cross-correlation. The default is False.
    margin : int, optional
        The margin of the region to compare, the default is 0.
    use_rfft : bool, optional
        Whether to use the `rfft` transform in place of the complex `fft` transform. The default is True.
    stack_axis : int, optional
        The axis along which the VU images are stacked. The default is -2.
    decimals : int, optional
        Decimals for the truncation of the sub-pixel  The default is 2.

    Returns
    -------
    NDArrayFloat
        The VU shifts.
    """
    num_angles = data_vwu.shape[stack_axis]

    if use_rfft:
        local_fftn = np.fft.rfftn
        local_ifftn = np.fft.irfftn
    else:
        local_fftn = np.fft.fftn
        local_ifftn = np.fft.ifftn

    fft_dims = np.delete(np.arange(-len(data_vwu.shape), 0), stack_axis)
    u_axis = fft_dims[-1]

    old_fft_shapes = np.array(np.array(data_vwu.shape)[fft_dims], ndmin=1, dtype=int)
    new_fft_shapes = old_fft_shapes.copy()
    if pad_u:
        new_fft_shapes[u_axis] *= 2
    cc_coords = [np.fft.fftfreq(s, 1 / s) for s in new_fft_shapes]

    if margin > 0:
        mask = np.zeros([proj_vwu.shape[d] for d in fft_dims], dtype=proj_vwu.dtype)
        slices = [slice(margin, proj_vwu.shape[d] - margin) for d in fft_dims]
        mask[tuple(slices)] = 1.0
        proj_vwu = proj_vwu * mask[..., None, :]

    if len(fft_dims) == 2:
        shifts_vu = np.empty((len(data_vwu.shape) - 1, num_angles))
        slices = [slice(None)] * len(data_vwu.shape)

        for ii_a in range(num_angles):
            # For performance reasons, it is better to do the fft on each image
            slices[stack_axis] = slice(ii_a, ii_a + 1)
            data_vu = data_vwu[tuple(slices)].squeeze(axis=stack_axis)
            if proj_vwu.shape[stack_axis] == 1:
                proj_vu = proj_vwu.squeeze(axis=stack_axis)
            else:
                proj_vu = proj_vwu[tuple(slices)].squeeze(axis=stack_axis)
            data_vwu_f = local_fftn(data_vu, s=list(new_fft_shapes))
            proj_vwu_f = local_fftn(proj_vu, s=list(new_fft_shapes))

            cc_f = data_vwu_f * proj_vwu_f.conj()
            if normalize_fourier:
                cc_f /= np.fmax(np.abs(cc_f), eps)
            cc_r: NDArrayFloat = local_ifftn(cc_f).real

            f_vals, f_coords = extract_peak_region_nd(cc_r, cc_coords=cc_coords)
            shifts_vu[..., ii_a] = np.array([f_coords[0][1], f_coords[1][1]])

            if decimals > 0:
                f_vals_v = f_vals[:, 1]
                f_vals_u = f_vals[1, :]

                sub_pixel_v = refine_max_position_1d(f_vals_v, decimals=decimals)
                sub_pixel_u = refine_max_position_1d(f_vals_u, decimals=decimals)

                shifts_vu[..., ii_a] += [sub_pixel_v, sub_pixel_u]
    else:
        data_vwu_f = local_fftn(data_vwu, s=list(new_fft_shapes), axes=list(fft_dims))
        proj_vwu_f = local_fftn(proj_vwu, s=list(new_fft_shapes), axes=list(fft_dims))

        ccs_f = data_vwu_f * proj_vwu_f.conj()
        if normalize_fourier:
            ccs_f /= np.fmax(np.abs(ccs_f).max(axis=u_axis, keepdims=True), eps)
        ccs = local_ifftn(ccs_f, axes=fft_dims).real

        f_vals, f_h = extract_peak_regions_1d(ccs, axis=u_axis, cc_coords=cc_coords[u_axis])
        shifts_vu = f_h[1, :]

        if decimals > 0:
            shifts_vu += refine_max_position_1d(f_vals, decimals=decimals)

    # import skimage.registration as skr
    # upsample_factor = int(1 / 10 ** (-decimals))
    # shifts_vu = np.empty((len(data_vwu.shape) - 1, num_angles))
    # for ii in range(num_angles):
    #     shifts_vu[..., ii] = skr.phase_cross_correlation(
    #         data_vwu[..., ii, :], proj_vwu[..., ii, :], upsample_factor=upsample_factor, return_error=False
    #     )

    return shifts_vu


def fit_shifts_zyx_xc(
    ref_vol_zyx: NDArrayFloat,
    rec_vol_zyx: NDArrayFloat,
    pad_zyx: bool = False,
    normalize_fourier: bool = True,
    use_rfft: bool = True,
    decimals: int = 2,
) -> NDArrayFloat:
    """
    Find the ZYX shifts of the volume, through cross-correlation.

    Parameters
    ----------
    ref_vol_zyx : NDArrayFloat
        The reference volume.
    rec_vol_zyx : NDArrayFloat
        The reconstructed volume to register.
    pad_zyx : bool, optional
        Pad the ZYX coordinates. The default is False.
    normalize_fourier : bool, optional
        Whether to normalize the Fourier representation of the cross-correlation. The default is True.
    use_rfft : bool, optional
        Whether to use the `rfft` transform in place of the complex `fft` transform. The default is True.
    decimals : int, optional
        Decimals for the truncation of the sub-pixel  The default is 2.

    Returns
    -------
    NDArrayFloat
        The ZYX shifts.
    """
    if use_rfft:
        local_fftn = np.fft.rfftn
        local_ifftn = np.fft.irfftn
    else:
        local_fftn = np.fft.fftn
        local_ifftn = np.fft.ifftn

    fft_dims = np.arange(-np.fmin(ref_vol_zyx.ndim, 3), 0)
    old_fft_shapes = np.array(np.array(ref_vol_zyx.shape)[fft_dims], ndmin=1, dtype=int)
    new_fft_shapes = old_fft_shapes.copy()
    if pad_zyx:
        new_fft_shapes *= 2
    cc_coords = [np.fft.fftfreq(s, 1 / s) for s in new_fft_shapes]

    ref_vol_zyx_f = local_fftn(ref_vol_zyx, s=list(new_fft_shapes), axes=fft_dims)
    rec_vol_zyx_f = local_fftn(rec_vol_zyx, s=list(new_fft_shapes), axes=fft_dims)

    cc_f = ref_vol_zyx_f * rec_vol_zyx_f.conj()
    if normalize_fourier:
        cc_f /= np.fmax(np.abs(cc_f), eps)
    cc: NDArrayFloat = local_ifftn(cc_f).real

    f_vals, f_coords = extract_peak_region_nd(cc, cc_coords=cc_coords)
    shifts_zyx = np.array([coords[1] for coords in f_coords])

    if decimals > 0:
        for ii, dim in enumerate(fft_dims):
            slices = [slice(1, 2)] * ref_vol_zyx.ndim
            slices[dim] = slice(None)
            f_vals_slice = f_vals[tuple(slices)].flatten()

            sub_pixel_pos = refine_max_position_1d(f_vals_slice, decimals=decimals)

            shifts_zyx[ii] += sub_pixel_pos

    return shifts_zyx


def fit_image_rotation_and_scale(
    img_1_vu: NDArray, img_2_vu: NDArray, pad_mode: Union[str, None] = None, window_type: str = "hann", verbose: bool = False
) -> tuple[float, float]:
    """Fit the rotation and scaling of an image against a reference image. This works best for larger rotation angles.

    Parameters
    ----------
    img_1_vu : NDArray
        Reference image
    img_2_vu : NDArray
        Rotated and scaled image
    pad_mode : Union[str, None], optional
        Padding mode, by default None
    window_type : str, optional
        Windowing type (to cud the high frequency aliasing), by default "hann"
    verbose : bool, optional
        Whether to give verbose output, by default False

    Returns
    -------
    tuple[float, float]
        The rotation (in degrees) and scale of the second image with respect to the first

    Raises
    ------
    ValueError
        In case of mismatching shape of the two images.
    """
    if img_1_vu.ndim != img_2_vu.ndim or np.any(np.array(img_1_vu.shape) != np.array(img_2_vu.shape)):
        raise ValueError(
            f"Image shapes should be identical, but instead got image #1: {img_1_vu.shape}, and image #2: {img_2_vu.shape}"
        )

    axes = (-2, -1)

    img_shape = img_2_vu.shape
    if pad_mode is not None:
        pad_widths = [(s // 2,) for s in img_shape]
        img_1_vu = np.pad(img_1_vu, pad_width=pad_widths, mode=pad_mode)
        img_2_vu = np.pad(img_2_vu, pad_width=pad_widths, mode=pad_mode)
        img_shape = img_2_vu.shape

    img_win = window(window_type=window_type, shape=img_shape)

    img_fft_1 = np.fft.fft2(img_1_vu * img_win, axes=axes)
    img_fft_2 = np.fft.fft2(img_2_vu * img_win, axes=axes)

    # abs removes the translation component
    img_fft_1 = np.abs(np.fft.fftshift(img_fft_1, axes=axes))
    img_fft_2 = np.abs(np.fft.fftshift(img_fft_2, axes=axes))

    # transform to polar coordinates
    img_center = [s - s // 2 for s in img_shape]
    radius = min([s // 2 for s in img_shape])
    img_fft_1_p = warp_polar(img_fft_1, center=img_center, scaling="log", radius=radius)
    img_fft_2_p = warp_polar(img_fft_2, center=img_center, scaling="log", radius=radius)

    # only use half of FFT
    img_fft_1_p = img_fft_1_p[..., : img_fft_1_p.shape[0] // 2, :]
    img_fft_2_p = img_fft_2_p[..., : img_fft_2_p.shape[0] // 2, :]

    fft_polar_shifts_rs = fit_shifts_vu_xc(img_fft_1_p[:, None, 1:], img_fft_2_p[:, None, 1:], normalize_fourier=True)
    tilt_pix = np.squeeze(fft_polar_shifts_rs[0])
    tilt_deg = (180 / img_fft_2_p.shape[0]) * tilt_pix

    klog = img_fft_2_p.shape[1] / np.log(radius)
    scale = np.exp(np.squeeze(fft_polar_shifts_rs[1]) / klog)

    if verbose:
        print(f"Fitted image rotation: {tilt_deg:.6} (degrees) or {tilt_pix} (pixels), with scale factor: {scale:.6}")

    return tilt_deg, scale


def fit_camera_tilt_angle(img_1: NDArray, img_2: NDArray, pad_u: bool = False, fit_l1: bool = True, verbose: bool = False):
    """
    Estimate the camera tilt angle based on correlation peak values between two images.

    Parameters
    ----------
    img_1: NDArray
        The first image.
    img_2: NDArray
        The second image.
    pad_u: bool, optional
        Enable zero padding. Default is False.
    fit_l1: bool, optional
        Perform L1 norm fitting if True. Default is True.
    verbose: bool, optional
        Enable verbose output. Default is False.

    Returns
    -------
    tuple[float, float]
        Tuple containing the estimated center of rotation offset (pixels) and camera tilt angle (degrees).
    """
    fitted_shifts_h = fit_shifts_vu_xc(img_1, img_2, pad_u=pad_u)
    fitted_cors = fitted_shifts_h / 2

    # Computing tilt
    img_shape = img_2.shape
    half_img_size = (img_shape[-2] - 1) / 2
    cc_v_coords = np.linspace(-half_img_size, half_img_size, img_shape[-2])
    poly_slope = Polynomial.fit(cc_v_coords, fitted_cors, deg=1)
    b, a = poly_slope.convert().coef

    if fit_l1:

        def f(coeffs: NDArray) -> float:
            b, a = coeffs[0], coeffs[1]
            pred_line = cc_v_coords * a + b
            l1_diff = np.linalg.norm(pred_line - fitted_cors, ord=1)
            return float(l1_diff)

        coeffs_opt = minimize(f, np.array([b, a]))
        b, a = coeffs_opt.x

    tilt_deg = np.rad2deg(-a / 2)
    cor_offset_pix = b

    if verbose:
        cor_trend = Polynomial([b, a])
        print(f"Fitted center of rotation (pixels): {cor_offset_pix}, and camera tilt (degrees): {tilt_deg}")
        fig, axs = plt.subplots(1, 1)
        axs.scatter(cc_v_coords, fitted_cors, label="Line CoRs")
        axs.plot(cc_v_coords, cor_trend(cc_v_coords), "-C1", label="Line CoRs trend")
        axs.axhline(cor_offset_pix, color="C2", linestyle="--", label=f"Image CoR ({cor_offset_pix:.3})")
        axs.set_title("Correlation peaks")
        axs.grid()
        axs.legend(fontsize=13)
        fig.tight_layout()
        plt.show(block=False)

    return cor_offset_pix, tilt_deg


def sinusoid(
    x: Union[NDArrayFloat, float], a: Union[NDArrayFloat, float], p: Union[NDArrayFloat, float], b: Union[NDArrayFloat, float]
) -> NDArrayFloat:
    """Compute the values of a sine function.

    Parameters
    ----------
    x : NDArrayFloat | float
        The independent variable.
    a : NDArrayFloat | float
        The amplitude of the sine.
    p : NDArrayFloat | float
        The phase of the sine.
    b : NDArrayFloat | float
        The bias of the sine.

    Returns
    -------
    NDArrayFloat
        The computed values.
    """
    return a * np.sin(x + p) + b


def fit_sinusoid(angles: NDArrayFloat, values: NDArrayFloat, fit_l1: bool = False) -> tuple[float, float, float]:
    """Fits a sinusoid to the given values.

    Parameters
    ----------
    angles : NDArrayFloat
        Angles where to evaluate the sinusoid.
    values : NDArrayFloat
        Values of the sinusoid.
    fit_l1 : bool, optional
        Whether to use l1 fit instead of the l2 fit, by default False

    Returns
    -------
    Tuple[float, float, float]
        The amplitude, phase and bias of the sinusoid.
    """
    a0 = (values.max() - values.min()) / 2
    b0 = (values.max() + values.min()) / 2

    (a, p, b), _ = spopt.curve_fit(sinusoid, angles, values, p0=[a0, 0, b0])

    if fit_l1:

        def f(apb: NDArrayFloat) -> float:
            a, p, b = apb[0], apb[1], apb[2]
            pred_sinusoid = sinusoid(angles, a, p, b)
            l1_diff = np.linalg.norm(pred_sinusoid - values, ord=1)
            return float(l1_diff)

        apb = spopt.minimize(f, np.array([a, p, b]))
        (a, p, b) = apb.x

    return a, p, b


def extract_peak_regions_1d(
    cc: NDArrayFloat, axis: int = -1, peak_radius: int = 1, cc_coords: Union[ArrayLike, NDArray, None] = None
) -> tuple[NDArrayFloat, Optional[NDArray]]:
    """
    Extract a region around the maximum value.

    Parameters
    ----------
    cc: NDArrayFloat
        Correlation image.
    axis: int, optional
        Find the max values along the specified direction. The default is -1.
    peak_radius: int, optional
        The l_inf radius of the area to extract around the peak. The default is 1.
    cc_coords: ArrayLike, optional
        The coordinates of `cc` along the selected axis. The default is None.

    Returns
    -------
    f_vals: NDArrayFloat
        The extracted function values.
    fc_ax: NDArrayFloat
        The coordinates of the extracted values, along the selected axis.
    """
    if len(cc.shape) == 1:
        cc = cc[None, ...]
    img_shape = np.array(cc.shape)
    if len(img_shape) != 2:
        raise ValueError(f"The input image should be either a 1 or 2-dimensional array. Array of shape: {cc.shape} was given.")
    other_axis = (axis + 1) % 2
    # get pixel having the maximum value of the correlation array
    pix_max = np.argmax(cc, axis=axis)

    # select a n neighborhood for the many 1D sub-pixel fittings (with wrapping)
    p_ax_range = np.arange(-peak_radius, peak_radius + 1)
    p_ax = (pix_max[None, :] + p_ax_range[:, None]) % img_shape[axis]

    p_ln = np.tile(np.arange(0, img_shape[other_axis])[None, :], [2 * peak_radius + 1, 1])

    # extract the pixel coordinates along the axis
    if cc_coords is None:
        fc_ax = None
    else:
        cc_coords = np.array(cc_coords, ndmin=1)
        fc_ax = cc_coords[p_ax.flatten()].reshape(p_ax.shape)

    # extract the correlation values
    if other_axis == 0:
        f_vals = cc[p_ln, p_ax]
    else:
        f_vals = cc[p_ax, p_ln]

    return (f_vals, fc_ax)


def refine_max_position_1d(
    f_vals: NDArrayFloat, f_x: Union[ArrayLike, NDArray, None] = None, return_vertex_val: bool = False, decimals: int = 2
) -> Union[NDArrayFloat, tuple[NDArrayFloat, NDArrayFloat]]:
    """Compute the sub-pixel max position of the given function sampling.

    Parameters
    ----------
    f_vals: NDArrayFloat
        Function values of the sampled points
    fx: ArrayLike, optional
        Coordinates of the sampled points
    return_vertex_val: boolean, option
        Enables returning the vertex values. Defaults to False.

    Raises
    ------
    ValueError
        In case position and values do not have the same size, or in case
        the fitted maximum is outside the fitting region.

    Returns
    -------
    float
        Estimated function max, according to the coordinates in fx.
    """
    if not len(f_vals.shape) in (1, 2):
        raise ValueError(
            f"The fitted values should be either one or a collection of 1-dimensional arrays."
            f" Array of shape: {f_vals.shape} was given."
        )
    num_vals = f_vals.shape[0]

    if f_x is None:
        f_x_half_size = (num_vals - 1) / 2
        f_x = np.linspace(-f_x_half_size, f_x_half_size, num_vals)
    else:
        f_x = np.squeeze(f_x)
        if not (len(f_x.shape) == 1 and np.all(f_x.size == num_vals)):
            raise ValueError(
                f"Base coordinates should have the same length as values array. Sizes of fx: {f_x.size}, f_vals: {num_vals}"
            )

    if len(f_vals.shape) == 1:
        # using Polynomial.fit, because supposed to be more numerically
        # stable than previous solutions (according to numpy).
        poly = Polynomial.fit(f_x, f_vals, deg=2)
        coeffs = poly.convert().coef
    else:
        coords = np.array([np.ones(num_vals), f_x, f_x**2])
        coeffs = np.linalg.lstsq(coords.T, f_vals, rcond=None)[0]

    # For a 1D parabola `f(x) = c + bx + ax^2`, the vertex position is:
    # x_v = -b / 2a.
    vertex_x = -coeffs[1, ...] / (2 * coeffs[2, ...])
    vertex_x = np.around(vertex_x, decimals=decimals)

    vertex_min_x = np.min(f_x)
    vertex_max_x = np.max(f_x)
    lower_bound_ok = vertex_min_x < vertex_x
    upper_bound_ok = vertex_x < vertex_max_x
    if not np.all(lower_bound_ok * upper_bound_ok):
        if len(f_vals.shape) == 1:
            message = (
                f"Fitted position {vertex_x} is outside the input margins [{vertex_min_x}, {vertex_max_x}]."
                f" Input values: {f_vals}"
            )
        else:
            message = (
                f"Fitted positions outside the input margins [{vertex_min_x}, {vertex_max_x}]:"
                f" {np.sum(1 - lower_bound_ok)} below and {np.sum(1 - upper_bound_ok)} above"
            )
        raise ValueError(message)

    if return_vertex_val:
        vertex_val = coeffs[0, ...] + vertex_x * coeffs[1, ...] / 2
        vertex_val = np.around(vertex_val, decimals=decimals)
        return vertex_x, vertex_val
    else:
        return vertex_x


def extract_peak_region_nd(
    cc: NDArrayFloat, peak_radius: int = 1, cc_coords: Optional[Sequence[Union[Sequence, NDArray]]] = None
) -> tuple[NDArray, Optional[Sequence[NDArray]]]:
    """
    Extract a region around the maximum value.

    Parameters
    ----------
    cc: NDArrayFloat
        Correlation image.
    peak_radius: int, optional
        The l_inf radius of the area to extract around the peak. The default is 1.
    cc_coords: ArrayLike, optional
        The coordinates of `cc`. The default is None.

    Returns
    -------
    f_vals: NDArrayFloat
        The extracted function values.
    f_coords: Sequence[NDArrayFloat] | None
        The coordinates of the extracted values.
    """
    img_shape = np.array(cc.shape)
    # get pixel having the maximum value of the correlation array
    pix_max_corr = np.argmax(cc)
    peak_pos = np.unravel_index(pix_max_corr, img_shape)

    # select a n x n neighborhood for the sub-pixel fitting (with wrapping)
    peak_ranges = [np.arange(p - peak_radius, p + peak_radius + 1) % img_shape[ii] for ii, p in enumerate(peak_pos)]

    # extract the (v, h) pixel coordinates
    if cc_coords is None:
        f_coords = None
    else:
        f_coords = [coords[pr] for coords, pr in zip(cc_coords, peak_ranges)]

    # extract the correlation values
    peak_ranges = np.meshgrid(*peak_ranges, indexing="ij")
    f_vals = cc[tuple(peak_ranges)]

    return f_vals, f_coords


def refine_max_position_2d(
    f_vals: NDArrayFloat, fy: Union[ArrayLike, NDArray, None] = None, fx: Union[ArrayLike, NDArray, None] = None
) -> NDArray:
    """Compute the sub-pixel max position of the given function sampling.

    Parameters
    ----------
    f_vals: NDArrayFloat
        Function values of the sampled points
    fy: ArrayLike, optional
        Vertical coordinates of the sampled points
    fx: ArrayLike, optional
        Horizontal coordinates of the sampled points

    Raises
    ------
    ValueError
        In case position and values do not have the same size, or in case
        the fitted maximum is outside the fitting region.

    Returns
    -------
    tuple(float, float)
        Estimated (vertical, horizontal) function max, according to the
        coordinates in fy and fx.
    """
    if not (len(f_vals.shape) == 2):
        raise ValueError(f"The fitted values should form a 2-dimensional array. Array of shape: {f_vals.shape} was given.")
    if fy is None:
        fy = np.linspace(-1, 1, f_vals.shape[0])
        y_scaling = (f_vals.shape[0] - 1) / 2
    else:
        fy = np.array(fy, ndmin=1)
        y_scaling = 1.0
        if not (len(fy.shape) == 1 and np.all(fy.size == f_vals.shape[0])):
            raise ValueError(
                f"Vertical coordinates should have the same length as values matrix."
                f" Sizes of fy: {fy.size}, f_vals: {f_vals.shape}"
            )

    if fx is None:
        fx = np.linspace(-1, 1, f_vals.shape[1])
        x_scaling = (f_vals.shape[1] - 1) / 2
    else:
        fx = np.array(fx, ndmin=1)
        x_scaling = 1.0
    if not (len(fx.shape) == 1 and np.all(fx.size == f_vals.shape[1])):
        raise ValueError(
            f"Horizontal coordinates should have the same length as values matrix."
            f"Sizes of fx: {fx.size}, f_vals: {f_vals.shape}"
        )

    fy, fx = np.meshgrid(fy, fx, indexing="ij")
    fy = fy.flatten()
    fx = fx.flatten()

    coords = np.array([np.ones(f_vals.size), fy, fx, fy * fx, fy**2, fx**2])
    coeffs = np.linalg.lstsq(coords.T, f_vals.flatten(), rcond=None)[0]
    coeffs *= [1, y_scaling, x_scaling, y_scaling * x_scaling, y_scaling**2, x_scaling**2]

    # For a 1D parabola `f(x) = ax^2 + bx + c`, the vertex position is:
    # x_v = -b / 2a. For a 2D parabola, the vertex position is:
    # (y, x)_v = - b / A, where:
    A = [[2 * coeffs[4], coeffs[3]], [coeffs[3], 2 * coeffs[5]]]
    b = coeffs[1:3]
    vertex_yx = np.linalg.lstsq(A, -b, rcond=None)[0]

    vertex_min_yx = [np.min(fy), np.min(fx)]
    vertex_max_yx = [np.max(fy), np.max(fx)]
    if np.any(vertex_yx < vertex_min_yx) or np.any(vertex_yx > vertex_max_yx):
        print(f_vals)
        raise ValueError(
            f"Fitted (yx: {vertex_yx}) positions are outside the input margins"
            + f" y: [{vertex_min_yx[0]}, {vertex_max_yx[0]}], and x: [{vertex_min_yx[1]}, {vertex_max_yx[1]}]."
            + f" Input values: {f_vals}"
        )
    return vertex_yx
